"use strict";
/**
 * Created by rburson on 8/29/17.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const client_1 = require("./client");
const util_1 = require("./util");
const models_1 = require("./models");
const ws_1 = require("./ws");
const persistence_1 = require("./persistence");
// Chose the locales to load based on this list:
// https://stackoverflow.com/questions/9711066/most-common-locales-for-worldwide-compatibility
// Best effort for now.  Need to dynamically load these from Globalize???
require("moment/locale/zh-cn");
require("moment/locale/ru");
require("moment/locale/fr");
require("moment/locale/es");
require("moment/locale/en-gb");
require("moment/locale/de");
require("moment/locale/pt");
require("moment/locale/pt-br");
require("moment/locale/en-ca");
require("moment/locale/it");
require("moment/locale/ja");
/**
 * Top-level entry point into the Catavolt API
 */
class CatavoltApi {
    /**
     * Construct an CatavoltApi
     * This should not be called directly, instead use the 'singleton' method
     * @private
     */
    constructor(serverUrl, serverVersion) {
        this.dataLastChangedTime = new Date(0);
        this._clientMode = client_1.ClientMode.ONLINE;
        if (CatavoltApi._singleton) {
            throw new Error("Singleton instance already created");
        }
        this._devicePropsStatic = {};
        this._devicePropsDynamic = {};
        this.initPersistentApi(serverUrl, serverVersion);
        CatavoltApi._singleton = this;
    }
    /* ********************
            Statics
     *********************** */
    /**
     * Get the default session time
     * @returns {number}
     */
    static get defaultTTLInMillis() {
        return CatavoltApi.ONE_HOUR_IN_MILLIS;
    }
    /**
     * Get the singleton instance of the CatavoltApi
     * @returns {CatavoltApi}
     */
    static get singleton() {
        if (!CatavoltApi._singleton) {
            CatavoltApi._singleton = new CatavoltApi(CatavoltApi.SERVER_URL, CatavoltApi.SERVER_VERSION);
        }
        return CatavoltApi._singleton;
    }
    /* *****************
        Public Ops
       ******************* */
    /**
     * Add or replace a dynamic device property (func)
     * @param propName
     * @param propFn
     */
    addDynamicDeviceProp(propName, propFn) {
        this._devicePropsDynamic[propName] = propFn;
    }
    /**
     * Add or replace a static device property
     *
     * @param propName
     * @param propValue
     */
    addStaticDeviceProp(propName, propValue) {
        this._devicePropsStatic[propName] = propValue;
    }
    /**
     * Get the json representation of this client's locale.  The server pulls this value from the agent string
     * and returns it to the client.
     * @returns {string}
     */
    get browserLocaleJson() {
        return this.session.tenantProperties['browserLocale'];
    }
    /*@TODO*/
    changePasswordAndLogin(tenantId, clientType, userId, existingPassword, newPassword) {
        return Promise.reject(new Error('Not Yet Implemented'));
    }
    /**
     * Get the number of millis that the client will remain active between calls
     * to the server.
     * @returns {number}
     */
    get clientTimeoutMillis() {
        const mins = this.session.tenantProperties['clientTimeoutMinutes'];
        return mins ? (Number(mins) * 60 * 1000) : CatavoltApi.defaultTTLInMillis;
    }
    /**
     * Get the currency symbol override if defined from the server.
     * @returns {string}
     */
    get currencySymbol() {
        const currencySymbol = this.session.tenantProperties['currencySymbol'];
        return currencySymbol ? currencySymbol : null;
    }
    /**
     * Get the device props
     * @returns {{[p: string]: string}}
     */
    get deviceProps() {
        const newProps = util_1.ObjUtil.addAllProps(this._devicePropsStatic, {});
        for (const attr in this._devicePropsDynamic) {
            newProps[attr] = this._devicePropsDynamic[attr]();
        }
        return newProps;
    }
    /**
     * Get the DialogApi instance
     * @returns {DialogApi}
     */
    get dialogApi() {
        return this._dialogApi;
    }
    /**
     * Get a Workbench by workbenchId
     * @param workbenchId
     * @returns {Promise<Workbench>}
     */
    getWorkbench(workbenchId) {
        if (!this.isLoggedIn) {
            return Promise.reject(new Error('User is not logged in'));
        }
        return this.dialogApi.getWorkbench(this.session.tenantId, this.session.id, workbenchId);
    }
    /**
     * Get the list of Workbenches
     *
     * @returns {Array<Workbench>}
     */
    getWorkbenches() {
        if (!this.isLoggedIn) {
            return Promise.reject(new Error('User is not logged in'));
        }
        return this.dialogApi.getWorkbenches(this.session.tenantId, this.session.id);
    }
    /**
     * Initialize a dialog service implementation for use by this CatavoltApi
     *
     * @param serverVersion
     * @param serverUrl
     */
    initDialogApi(serverUrl, serverVersion = CatavoltApi.SERVER_VERSION) {
        this._dialogApi = new DialogService(new ws_1.FetchClient(this._clientMode), serverUrl, serverVersion);
    }
    /**
     * Initialize an offline dialog service
     *
     * @param serverVersion
     * @param serverUrl
     */
    initPersistentApi(serverUrl, serverVersion) {
        this._dialogApi = new DialogService(new persistence_1.PersistentClient(this._clientMode), serverUrl, serverVersion);
    }
    isOfflineMode() {
        return this._clientMode === client_1.ClientMode.OFFLINE;
    }
    isPersistentClient() {
        return this.dialogApi.client instanceof persistence_1.PersistentClient;
    }
    /**
     * Check for the availability of the given featureSet
     * @see FeatureSet
     * @param featureSet
     * @returns {boolean}
     */
    isFeatureSetAvailable(featureSet) {
        try {
            const currentVersion = AppVersion.getAppVersion(this.session.serverVersion);
            const featureMinimumVersion = FeatureVersionMap[featureSet];
            return featureMinimumVersion.isLessThanOrEqualTo(currentVersion);
        }
        catch (error) {
            util_1.Log.error('Failed to compare appVersions for feature ' + featureSet);
            util_1.Log.error(error);
            return false;
        }
    }
    /**
     * Checked logged in status
     * @returns {boolean}
     */
    get isLoggedIn() {
        return !!this._session;
    }
    /**
     * Log in and retrieve the Session
     *
     * @param tenantId
     * @param clientType
     * @param userId
     * @param password
     * @param serverUrl
     * @param serverVersion
     *
     * @returns {Promise<Session | Redirection>}
     */
    login(tenantId, clientType, userId, password) {
        if (this.isLoggedIn) {
            return Promise.reject(new Error('User is already logged in'));
        }
        const login = {
            userId: userId,
            password: password,
            clientType: clientType,
            deviceProperties: this.deviceProps,
            type: models_1.TypeNames.LoginTypeName
        };
        return this.dialogApi.createSession(tenantId, login).then((result) => {
            if (result.type === models_1.TypeNames.SessionTypeName) {
                this._session = result;
                return result;
            }
            else {
                return result;
            }
        });
    }
    /**
     * Logout and destroy the session
     * @returns {{sessionId:string}}
     */
    logout() {
        if (!this.isLoggedIn) {
            return Promise.reject('User is already logged out');
        }
        return this.dialogApi.deleteSession(this.session.tenantId, this.session.id).then(result => {
            this._session = null;
            return result;
        });
    }
    openDialog(redirection) {
        return this.dialogApi.getDialog(this.session.tenantId, this.session.id, redirection.dialogId)
            .then((dialog) => {
            dialog.initialize(this);
            if (dialog.view instanceof models_1.Form) {
                return dialog;
            }
            else {
                throw new Error(`Unexpected top-level dialog view type: ${dialog.view.type}`);
            }
        });
    }
    toDialogOrRedirection(resultPr) {
        return resultPr.then((actionResult) => {
            if (models_1.RedirectionUtil.isDialogRedirection(actionResult)) {
                return this.openDialog(actionResult);
            }
            else if (models_1.RedirectionUtil.isRedirection(actionResult)) {
                return Promise.resolve(actionResult);
            }
            else {
                //@TODO - this shouldn't be a null redirection - what should it be?
                return Promise.resolve(actionResult);
            }
        });
    }
    getRedirection(redirectionId) {
        return this.dialogApi.getRedirection(this.session.tenantId, this.session.id, redirectionId);
    }
    /**
     * Open a {@link WorkbenchAction}
     * @param workbenchAction
     * @returns {Promise<{actionId:string} | Redirection>}
     */
    performWorkbenchAction(workbenchAction) {
        return this.performWorkbenchActionForId(workbenchAction.workbenchId, workbenchAction.id);
    }
    /**
     * Open a {@link WorkbenchWorkbenchAction}
     * @param workbenchId
     * @param workbenchActionId
     * @returns {Promise<{actionId:string} | Redirection>}
     */
    performWorkbenchActionForId(workbenchId, workbenchActionId) {
        if (!this.isLoggedIn) {
            return Promise.reject(new Error('User is not logged in'));
        }
        return this.dialogApi.performWorkbenchAction(this.session.tenantId, this.session.id, workbenchId, workbenchActionId);
    }
    /**
     * Refresh the CatavoltApi
     *
     * @returns {Promise<Session>}
     */
    refreshSession(tenantId, sessionId) {
        return this.dialogApi.getSession(tenantId, sessionId).then(session => {
            this._session = session;
            return session;
        });
    }
    /**
     * Time remaining before this session is expired by the server
     * @returns {number}
     */
    get remainingSessionTime() {
        return this.clientTimeoutMillis - ((new Date()).getTime() - this.dialogApi.lastServiceActivity.getTime());
    }
    /**
     * Get the Session
     * @returns {Session}
     */
    get session() {
        return this._session;
    }
    /**
     * Return whether or not the session has expired
     * @returns {boolean}
     */
    get sessionHasExpired() {
        return this.remainingSessionTime < 0;
    }
    setPersistentClient() {
        this.initPersistentApi(CatavoltApi.SERVER_URL, CatavoltApi.SERVER_VERSION);
    }
    setOnlineClient() {
        this.initDialogApi(CatavoltApi.SERVER_URL, CatavoltApi.SERVER_VERSION);
    }
    setOfflineMode() {
        this._clientMode = client_1.ClientMode.OFFLINE;
        this.dialogApi.setClientMode(this._clientMode);
    }
    setOnlineMode() {
        this._clientMode = client_1.ClientMode.ONLINE;
        this.dialogApi.setClientMode(this._clientMode);
    }
}
CatavoltApi.ONE_HOUR_IN_MILLIS = 60 * 60 * 1000;
//defaults
CatavoltApi.SERVER_URL = 'https://dialog.hxgn-api.net';
CatavoltApi.SERVER_VERSION = 'v0';
exports.CatavoltApi = CatavoltApi;
class DialogService {
    constructor(client, serverUrl = DialogService.SERVER, apiVersion = 'v0') {
        this.client = client;
        this.apiVersion = apiVersion;
        this.baseUrl = `${serverUrl}/${apiVersion}`;
    }
    createSession(tenantId, login) {
        return this.post(`tenants/${tenantId}/sessions`, login).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValueOrRedirect());
    }
    getSession(tenantId, sessionId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    deleteSession(tenantId, sessionId) {
        return this.d3lete(`tenants/${tenantId}/sessions/${sessionId}`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    getWorkbenches(tenantId, sessionId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/workbenches`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    getWorkbench(tenantId, sessionId, workbenchId) {
        return this.get(`tenants/{$tenantId}/sessions/{$sessionId}/workbenches/{$workbenchId}`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    getRedirection(tenantId, sessionId, redirectionId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/redirections/${redirectionId}`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    getDialog(tenantId, sessionId, dialogId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    deleteDialog(tenantId, sessionId, dialogId) {
        return this.d3lete(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    getActions(tenantId, sessionId, dialogId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/actions`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    performAction(tenantId, sessionId, dialogId, actionId, actionParameters) {
        const encodedActionId = encodeURIComponent(actionId);
        return this.post(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/actions/${encodedActionId}`, actionParameters).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValueOrRedirect());
    }
    getWorkbenchActions(tenantId, sessionId, workbenchId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/workbenches/${workbenchId}/actions`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    performWorkbenchAction(tenantId, sessionId, workbenchId, actionId) {
        return this.post(`tenants/${tenantId}/sessions/${sessionId}/workbenches/${workbenchId}/actions/${actionId}`, {}).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValueOrRedirect());
    }
    getRecord(tenantId, sessionId, dialogId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/record`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    putRecord(tenantId, sessionId, dialogId, record) {
        return this.put(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/record`, record).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValueOrRedirect());
    }
    getRecords(tenantId, sessionId, dialogId, queryParams) {
        return this.post(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/records`, queryParams).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    getMode(tenantId, sessionId, dialogId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/viewMode`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    changeMode(tenantId, sessionId, dialogId, mode) {
        return this.put(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/viewMode/${mode}`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    getView(tenantId, sessionId, dialogId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/view`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    changeView(tenantId, sessionId, dialogId, viewId) {
        return this.put(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/selectedView/{viewId}`, {}).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    getViews(tenantId, sessionId, dialogId) {
        return this.get(`tenants/${tenantId}/sessions/${sessionId}/dialogs/${dialogId}/availableViews`).then(jsonClientResponse => (new DialogServiceResponse(jsonClientResponse)).responseValue());
    }
    get lastServiceActivity() {
        return this.client.lastActivity;
    }
    setClientMode(clientMode) {
        this.client.setClientMode(clientMode);
    }
    /* Private methods */
    get(path, queryParams) {
        return this.client.getJson(`${DialogService.SERVER}/${this.apiVersion}`, path, queryParams);
    }
    post(path, body) {
        return this.client.postJson(`${DialogService.SERVER}/${this.apiVersion}`, path, body);
    }
    d3lete(path) {
        return this.client.deleteJson(`${DialogService.SERVER}/${this.apiVersion}`, path);
    }
    put(path, body) {
        return this.client.putJson(`${DialogService.SERVER}/${this.apiVersion}`, path, body);
    }
}
DialogService.SERVER = 'https://dialog.hxgn-api.net';
exports.DialogService = DialogService;
class DialogServiceResponse {
    constructor(clientResponse) {
        this.clientResponse = clientResponse;
    }
    responseValue() {
        return new Promise((resolve, reject) => {
            if (this.hasError) {
                reject(this.clientResponse.value);
            }
            else {
                this.fullfillJsonToModel(this.clientResponse, resolve, reject);
            }
        });
    }
    responseValueOrRedirect() {
        return new Promise((resolve, reject) => {
            if (this.hasError) {
                reject(this.clientResponse.value);
            }
            else if (this.hasValue) {
                this.fullfillJsonToModel(this.clientResponse, resolve, reject);
            }
            else {
                this.fullfillJsonToModel(this.clientResponse, resolve, reject);
            }
        });
    }
    assertNoError() {
        return new Promise((resolve, reject) => {
            if (this.hasError) {
                reject(this.clientResponse.value);
            }
            else {
                resolve(undefined);
            }
        });
    }
    get hasValue() {
        return this.clientResponse.statusCode >= 200 && this.clientResponse.statusCode < 300;
    }
    get hasRedirection() {
        return this.clientResponse.statusCode >= 300 && this.clientResponse.statusCode < 400;
    }
    get hasError() {
        return this.clientResponse.statusCode >= 400;
    }
    fullfillJsonToModel(clientResponse, resolve, reject) {
        models_1.ModelUtil.jsonToModel(this.clientResponse.value).then(resolve).catch(reject);
    }
}
exports.DialogServiceResponse = DialogServiceResponse;
/**
 *****************************************************
 */
/* Begin Feature Versioning */
class AppVersion {
    constructor(major, minor, patch) {
        this.major = major;
        this.minor = minor;
        this.patch = patch;
    }
    static getAppVersion(versionString) {
        const [major, minor, patch] = versionString.split('.');
        return new AppVersion(Number(major || 0), Number(minor || 0), Number(patch || 0));
    }
    /**
     * Is 'this' version less than or equal to the supplied version?
     * @param anotherVersion - the version to compare to 'this' version
     * @returns {boolean}
     */
    isLessThanOrEqualTo(anotherVersion) {
        if (anotherVersion.major > this.major) {
            return true;
        }
        else if (anotherVersion.major == this.major) {
            if (anotherVersion.minor > this.minor) {
                return true;
            }
            else if (anotherVersion.minor == this.minor) {
                return anotherVersion.patch >= this.patch;
            }
            else {
                return false;
            }
        }
        else {
            return false;
        }
    }
}
/* Map features to minimum app versions */
const FeatureVersionMap = {
    "View_Support": AppVersion.getAppVersion("1.3.447"),
    "Unified_Search": AppVersion.getAppVersion("1.3.463")
};
/* End Feature Versioning */
/*
    Export the Catavolt Object
 */
exports.Catavolt = CatavoltApi.singleton;
