"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const client_1 = require("./client");
const ws_1 = require("./ws");
const briefcase_template_1 = require("./briefcase-template");
const persistence_tools_1 = require("./persistence-tools");
class PersistentClient {
    constructor(clientMode = client_1.ClientMode.ONLINE) {
        /* Last operation happened at this time */
        this._lastActivity = new Date();
        this._clientMode = clientMode;
        this._fetchClient = new ws_1.FetchClient();
        const PersistentClientVars = JSON.parse(window.localStorage.getItem('PersistentClientVars'));
        if (PersistentClientVars) {
            this._tenantId = PersistentClientVars.tenantId;
            this._userId = PersistentClientVars.userId;
            this._sessionId = PersistentClientVars.sessionId;
            this._clientMode = PersistentClientVars.clientMode === 'OFFLINE' ? client_1.ClientMode.OFFLINE : client_1.ClientMode.ONLINE;
        }
    }
    get lastActivity() {
        return this._lastActivity;
    }
    writePersistentClientVars() {
        const PersistentClientVars = {
            tenantId: this._tenantId,
            userId: this._userId,
            sessionId: this._sessionId,
            clientMode: this._clientMode === client_1.ClientMode.OFFLINE ? 'OFFLINE' : 'ONLINE'
        };
        window.localStorage.setItem('PersistentClientVars', JSON.stringify(PersistentClientVars));
    }
    changeClientMode(clientMode) {
        this._clientMode = clientMode;
        this.writePersistentClientVars();
    }
    changeUserId(userId) {
        this._userId = userId;
        this.writePersistentClientVars();
    }
    changeSessionId(sessionId) {
        this._sessionId = sessionId;
        this.writePersistentClientVars();
    }
    changeTenantId(tenantId) {
        this._tenantId = tenantId;
        this.writePersistentClientVars();
    }
    createDialogMessageModel(message) {
        return { type: 'hxgn.api.dialog.DialogMessage', message: message };
    }
    createOfflineSessionNotFoundModel() {
        return this.createDialogMessageModel('Offline session not found');
    }
    createSessionIdModel(sessionId) {
        return { type: 'hxgn.api.dialog.SessionId', sessionId: sessionId };
    }
    deconstructGetDialogPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            dialogId: resourcePathElems[5]
        };
    }
    deconstructGetRecordPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            dialogId: resourcePathElems[5]
        };
    }
    deconstructGetRedirectionPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            redirectionId: resourcePathElems[5]
        };
    }
    deconstructPostMenuActionPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            dialogId: resourcePathElems[5],
            actionId: resourcePathElems[7]
        };
    }
    deconstructPostRecordsPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            dialogId: resourcePathElems[5]
        };
    }
    deconstructPostWorkbenchActionPath(resourcePathElems) {
        return {
            tenantId: resourcePathElems[1],
            sessionId: resourcePathElems[3],
            workbenchId: resourcePathElems[5],
            actionId: resourcePathElems[7]
        };
    }
    deleteJson(baseUrl, resourcePath) {
        const resourcePathElems = resourcePath.split('/');
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            if (persistence_tools_1.PersistenceTools.isDeleteSession(resourcePathElems)) {
                return this.deleteSession(baseUrl, resourcePath);
            }
            else {
                return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`DELETE action is not valid while offline: ${resourcePath}`), 400));
            }
        }
        return this._fetchClient.deleteJson(baseUrl, resourcePath);
    }
    deleteSession(baseUrl, resourcePath) {
        return new Promise((resolve, reject) => {
            const session = persistence_tools_1.PersistenceTools.readSessionState(this._tenantId, this._userId);
            if (!session) {
                resolve(new client_1.JsonClientResponse(this.createOfflineSessionNotFoundModel(), 404));
            }
            else {
                const sessionIdModel = this.createSessionIdModel(session.id);
                resolve(new client_1.JsonClientResponse(sessionIdModel, 200));
            }
        });
    }
    getBlob(baseUrl, resourcePath) {
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.BlobClientResponse(null, 400));
        }
        return this._fetchClient.getBlob(baseUrl, resourcePath);
        ;
    }
    getDialog(baseUrl, resourcePath, resourcePathElems, queryParams) {
        const pathFields = this.deconstructGetDialogPath(resourcePathElems);
        if (pathFields.dialogId === PersistentClient.BRIEFCASE_DIALOG_ID) {
            var dialog = persistence_tools_1.PersistenceTools.readDialogState(this._tenantId, this._userId, pathFields.dialogId);
            if (!dialog) {
                dialog = briefcase_template_1.BriefcaseTemplate.BRIEFCASE_DIALOG;
                dialog.sessionId = this._sessionId;
                persistence_tools_1.PersistenceTools.writeDialogState(this._tenantId, this._userId, dialog);
                persistence_tools_1.PersistenceTools.writeAllDialogParentState(this._tenantId, this._userId, dialog);
            }
            return Promise.resolve(new client_1.JsonClientResponse(dialog, 200));
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            const dialog = persistence_tools_1.PersistenceTools.readDialogState(this._tenantId, this._userId, pathFields.dialogId);
            if (!dialog) {
                return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`Offline dialog not found: ${pathFields.dialogId}`), 404));
            }
            else {
                return Promise.resolve(new client_1.JsonClientResponse(dialog, 200));
            }
        }
        let response = this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
        return response.then(jcr => {
            if (jcr.statusCode === 200) {
                var dialog = jcr.value;
                if (this.isWorkPackagesDialog(dialog)) {
                    dialog = this.patchWorkPackagesDialog(dialog);
                }
                // If a redirection object has been persisted, then also persist the dialog object.
                const redirection = persistence_tools_1.PersistenceTools.readRedirectionState(this._tenantId, this._userId, dialog.id);
                if (redirection) {
                    persistence_tools_1.PersistenceTools.writeDialogState(this._tenantId, this._userId, dialog);
                    persistence_tools_1.PersistenceTools.writeAllDialogParentState(this._tenantId, this._userId, dialog);
                }
                return new client_1.JsonClientResponse(dialog, 200);
            }
            return response;
        });
    }
    getJson(baseUrl, resourcePath, queryParams) {
        const resourcePathElems = resourcePath.split('/');
        if (persistence_tools_1.PersistenceTools.isGetDialog(resourcePathElems)) {
            return this.getDialog(baseUrl, resourcePath, resourcePathElems, queryParams);
        }
        else if (persistence_tools_1.PersistenceTools.isGetRedirection(resourcePathElems)) {
            return this.getRedirection(baseUrl, resourcePath, resourcePathElems, queryParams);
        }
        else if (persistence_tools_1.PersistenceTools.isGetSession(resourcePathElems)) {
            return this.getSession(baseUrl, resourcePath, resourcePathElems, queryParams);
        }
        else if (persistence_tools_1.PersistenceTools.isGetRecord(resourcePathElems)) {
            return this.getRecord(baseUrl, resourcePath, resourcePathElems, queryParams);
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`GET action is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
    }
    getRecord(baseUrl, resourcePath, resourcePathElems, queryParams) {
        const pathFields = this.deconstructGetRecordPath(resourcePathElems);
        if (pathFields.dialogId === PersistentClient.BRIEFCASE_RECORD_DIALOG_ID) {
            var record = persistence_tools_1.PersistenceTools.readRecordState(this._tenantId, this._userId, pathFields.dialogId);
            if (!record) {
                record = briefcase_template_1.BriefcaseTemplate.BRIEFCASE_RECORD;
                persistence_tools_1.PersistenceTools.writeRecordState(this._tenantId, this._userId, pathFields.dialogId, record);
            }
            return Promise.resolve(new client_1.JsonClientResponse(record, 200));
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            const record = persistence_tools_1.PersistenceTools.readRecordState(this._tenantId, this._userId, pathFields.dialogId);
            if (!record) {
                return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`Offline record not found: ${pathFields.dialogId}`), 404));
            }
            else {
                return Promise.resolve(new client_1.JsonClientResponse(record, 200));
            }
        }
        let response = this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
        return response.then(jcr => {
            if (jcr.statusCode === 200) {
                var record = jcr.value;
                const dialogId = pathFields.dialogId;
                var dialog = persistence_tools_1.PersistenceTools.findRootDialogState(this._tenantId, this._userId, dialogId);
                // If we have persisted the dialog, we also need to persist the dialog records
                if (dialog) {
                    persistence_tools_1.PersistenceTools.writeRecordState(this._tenantId, this._userId, dialogId, record);
                }
                return new client_1.JsonClientResponse(record, 200);
            }
            return response;
        });
    }
    getRedirection(baseUrl, resourcePath, resourcePathElems, queryParams) {
        const pathFields = this.deconstructGetRedirectionPath(resourcePathElems);
        if (pathFields.redirectionId === PersistentClient.BRIEFCASE_DIALOG_ID) {
            return Promise.resolve(new client_1.JsonClientResponse(briefcase_template_1.BriefcaseTemplate.BRIEFCASE_REDIRECTION, 303));
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`GET redirection is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
    }
    getSession(baseUrl, resourcePath, resourcePathElems, queryParams) {
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`GET session is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.getJson(baseUrl, resourcePath, queryParams);
    }
    getText(baseUrl, resourcePath) {
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.TextClientResponse(null, 400));
        }
        return this._fetchClient.getText(baseUrl, resourcePath);
    }
    isWorkPackagesDialog(dialog) {
        const referringObject = dialog.referringObject;
        return referringObject &&
            referringObject.type === PersistentClient.REFERRING_WORKBENCH_MODEL_TYPE &&
            referringObject.actionId === PersistentClient.WORK_PACKAGES_ACTION_ID;
    }
    patchWorkPackagesDialog(dialog) {
        const workPackagesTableDialog = dialog.children[0];
        const propertyDefs = workPackagesTableDialog.recordDef.propertyDefs;
        propertyDefs.push({
            "writeAllowed": false,
            "propertyName": "briefcase",
            "canCauseSideEffects": false,
            "upperCaseOnly": false,
            "propertyType": "boolean",
            "type": "hxgn.api.dialog.PropertyDef",
            "writeEnabled": false
        });
        const columns = workPackagesTableDialog.view.columns;
        columns.push({
            "propertyName": "briefcase",
            "heading": "Briefcase",
            "type": "hxgn.api.dialog.Column"
        });
        return dialog;
    }
    patchWorkPackagesRecordSet(dialog, recordSet) {
        const existingRecordSet = this.readBriefcaseWorkpackageRecordSet();
        const existingRecords = existingRecordSet ? existingRecordSet.records : [];
        const records = recordSet.records;
        if (records) {
            for (let r of records) {
                var existingBriefcase = false;
                for (let existingRecord of existingRecords) {
                    if (r.id === existingRecord.id) {
                        existingBriefcase = true;
                        break;
                    }
                }
                const briefcaseField = {
                    "name": "briefcase",
                    "annotations": [],
                    "type": "hxgn.api.dialog.Property",
                    "value": existingBriefcase
                };
                r.properties.push(briefcaseField);
            }
        }
        return recordSet;
    }
    makeNullRedirectionId() {
        return `null_redirection__offline_${Date.now()}`;
    }
    postAddToBriefcaseMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostMenuActionPath(resourcePathElems);
        const redirectionId = this.makeNullRedirectionId;
        if (!jsonBody || !jsonBody.targets || jsonBody.targets.length === 0) {
            return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel('Selection required'), 400));
        }
        const recordSet = persistence_tools_1.PersistenceTools.readRecordSetState(this._tenantId, this._userId, pathFields.dialogId);
        if (!recordSet) {
            return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel('Workpacakges not found'), 400));
        }
        const selectedRecords = [];
        for (let target of jsonBody.targets) {
            for (let record of recordSet.records) {
                if (record.id === target) {
                    persistence_tools_1.PersistenceTools.updateRecordPropertyValue(record, 'briefcase', true);
                    selectedRecords.push(record);
                }
            }
        }
        persistence_tools_1.PersistenceTools.writeRecordSetState(this._tenantId, this._userId, pathFields.dialogId, recordSet);
        const recordsAddedToBriefcase = [];
        for (let selected of selectedRecords) {
            const briefcaseProps = [];
            for (let p of selected.properties) {
                if (p.name === 'Disciplines') {
                    briefcaseProps.push({
                        "name": "disciplines",
                        "format": null,
                        "annotations": [],
                        "type": "hxgn.api.dialog.Property",
                        "value": p.value
                    });
                }
                else if (p.name === 'Owning_Group') {
                }
                else if (p.name === 'Description') {
                    briefcaseProps.push({
                        "name": "description",
                        "format": null,
                        "annotations": [],
                        "type": "hxgn.api.dialog.Property",
                        "value": p.value
                    });
                }
                else if (p.name === 'Config') {
                }
                else if (p.name === 'Name') {
                    briefcaseProps.push({
                        "name": "name",
                        "annotations": [],
                        "type": "hxgn.api.dialog.Property",
                        "value": p.value
                    });
                }
                else if (p.name === 'UID') {
                }
                else if (p.name === 'Creation_User') {
                }
                else if (p.name === 'Classification') {
                }
                else if (p.name === 'Organizations') {
                }
                else if (p.name === 'Creation_Date') {
                    briefcaseProps.push({
                        "name": "creation_date",
                        "format": "date",
                        "annotations": [],
                        "type": "hxgn.api.dialog.Property",
                        "value": p.value
                    });
                }
                else if (p.name === 'Last_Update_Date') {
                    briefcaseProps.push({
                        "name": "last_update_date",
                        "format": "date",
                        "annotations": [],
                        "type": "hxgn.api.dialog.Property",
                        "value": p.value
                    });
                }
                else if (p.name === 'Id') {
                    briefcaseProps.push({
                        "name": "workpackageid",
                        "annotations": [],
                        "type": "hxgn.api.dialog.Property",
                        "value": p.value
                    });
                }
                else if (p.name === 'Contract') {
                }
            }
            const nextAdd = {
                "annotations": [],
                "id": selected.id,
                "type": "hxgn.api.dialog.Record",
                "properties": briefcaseProps
            };
            recordsAddedToBriefcase.push(nextAdd);
        }
        const briefcaseWorkpackagesRecordSet = this.readBriefcaseWorkpackageRecordSet();
        briefcaseWorkpackagesRecordSet.records = briefcaseWorkpackagesRecordSet.records.concat(recordsAddedToBriefcase);
        this.writeBriefcaseWorkpackageRecordSet(briefcaseWorkpackagesRecordSet);
        const actionParameters = {
            "targets": [jsonBody.targets],
            "type": "hxgn.api.dialog.ActionParameters"
        };
        const nullRedirection = {
            "tenantId": this._tenantId,
            "referringObject": {
                "dialogMode": "LIST",
                "dialogProperties": {
                    "globalRefresh": "true",
                    "localRefresh": "true",
                    "dialogAlias": "Workpackage_AddToBriefcase"
                },
                "actionId": PersistentClient.ADD_TO_BRIEFCASE_ACTION_ID,
                "type": "hxgn.api.dialog.ReferringDialog",
                "dialogId": pathFields.actionId
            },
            "sessionId": pathFields.sessionId,
            "id": redirectionId,
            "type": "hxgn.api.dialog.NullRedirection"
        };
        return Promise.resolve(new client_1.JsonClientResponse(nullRedirection, 303));
    }
    postJson(baseUrl, resourcePath, jsonBody) {
        const resourcePathElems = resourcePath.split('/');
        if (persistence_tools_1.PersistenceTools.isPostSession(resourcePathElems)) {
            return this.postSession(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        else if (persistence_tools_1.PersistenceTools.isPostMenuAction(resourcePathElems)) {
            return this.postMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        else if (persistence_tools_1.PersistenceTools.isPostWorkbenchAction(resourcePathElems)) {
            return this.postWorkbenchAction(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        else if (persistence_tools_1.PersistenceTools.isPostRecords(resourcePathElems)) {
            return this.postRecords(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`POST action is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
    }
    postMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostMenuActionPath(resourcePathElems);
        if (pathFields.actionId === PersistentClient.ADD_TO_BRIEFCASE_ACTION_ID) {
            return this.postAddToBriefcaseMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        if (pathFields.actionId === PersistentClient.ENTER_OFFLINE_MODE_ACTION_ID || pathFields.actionId === PersistentClient.EXIT_OFFLINE_MODE_ACTION_ID) {
            const redirectionId = this.makeNullRedirectionId();
            const nullRedirection = {
                "tenantId": this._tenantId,
                "referringObject": {
                    "dialogMode": "READ",
                    "dialogProperties": {
                        "globalRefresh": "true",
                        "localRefresh": "true"
                    },
                    "actionId": pathFields.actionId,
                    "type": "hxgn.api.dialog.ReferringDialog",
                    "dialogId": pathFields.dialogId,
                },
                "sessionId": pathFields.sessionId,
                "id": redirectionId,
                "type": "hxgn.api.dialog.NullRedirection"
            };
            return new Promise((resolve, reject) => {
                var online = pathFields.actionId === PersistentClient.EXIT_OFFLINE_MODE_ACTION_ID;
                const briefcaseRecord = this.readBriefcaseRecord();
                if (!briefcaseRecord) {
                    resolve(new client_1.JsonClientResponse(this.createDialogMessageModel('Briefcase not found'), 400));
                }
                persistence_tools_1.PersistenceTools.updateRecordPropertyValue(briefcaseRecord, PersistentClient.ONLINE_PROPERTY_NAME, online);
                this.writeBriefcaseRecord(briefcaseRecord);
                this.changeClientMode(online ? client_1.ClientMode.ONLINE : client_1.ClientMode.OFFLINE);
                resolve(new client_1.JsonClientResponse(nullRedirection, 303));
            });
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            const target = jsonBody.targets[0];
            var alias = null;
            if (pathFields.actionId == 'alias_Open') {
                const referringAlias = persistence_tools_1.PersistenceTools.readDialogAliasState(this._tenantId, this._userId, pathFields.dialogId);
                alias = `Documents(${target})`;
                const navigationId = `${referringAlias}.${pathFields.actionId}.${alias}`;
                var navigation = persistence_tools_1.PersistenceTools.readNavigationState(this._tenantId, this._userId, navigationId);
                if (navigation) {
                    const redirection = persistence_tools_1.PersistenceTools.readRedirectionState(this._tenantId, this._userId, navigation.redirectionId);
                    if (redirection) {
                        return Promise.resolve(new client_1.JsonClientResponse(redirection, 303));
                    }
                    else {
                        return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`Documents redirection not found at: ${pathFields.dialogId}`), 404));
                    }
                }
                else {
                    return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`Documents navigation not found at: ${pathFields.dialogId}`), 404));
                }
            }
        }
        let response = this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
        return response.then(jcr => {
            if (jcr.statusCode === 303) {
                var redirection = jcr.value;
                if (redirection && redirection.referringObject && redirection.referringObject.type === PersistentClient.REFERRING_DIALOG_MODEL_TYPE) {
                    const target = jsonBody.targets[0];
                    const referringDialog = redirection.referringObject;
                    const rootDialog = persistence_tools_1.PersistenceTools.findRootDialogState(this._tenantId, this._userId, referringDialog.dialogId);
                    if (rootDialog) {
                        const referringAlias = persistence_tools_1.PersistenceTools.readDialogAliasState(this._tenantId, this._userId, referringDialog.dialogId);
                        var alias = null;
                        if (pathFields.actionId == 'alias_Open') {
                            alias = `Documents(${target})`;
                        }
                        else if (pathFields.actionId == 'alias_ShowTags') {
                            alias = `Tags(${target})`;
                        }
                        else if (pathFields.actionId == 'open') {
                            if (referringAlias.startsWith('Tags')) {
                                alias = `TagDetails(${target})`;
                            }
                            else if (referringAlias.startsWith('Documents')) {
                                alias = `DocumentDetails(${target})`;
                            }
                        }
                        const navigationId = alias ? `${referringAlias}.${pathFields.actionId}.${alias}` :
                            `${referringAlias}.${pathFields.actionId}.not_available`;
                        var navigation = persistence_tools_1.PersistenceTools.readNavigationState(this._tenantId, this._userId, navigationId);
                        if (navigation) {
                            persistence_tools_1.PersistenceTools.deleteAllDialogState(this._tenantId, this._userId, navigation.redirectionId);
                        }
                        navigation = {
                            id: navigationId,
                            redirectionId: redirection.id,
                            redirectionAlias: alias
                        };
                        persistence_tools_1.PersistenceTools.writeNavigationState(this._tenantId, this._userId, navigation);
                        persistence_tools_1.PersistenceTools.writeRedirectionState(this._tenantId, this._userId, redirection);
                        if (alias) {
                            persistence_tools_1.PersistenceTools.writeDialogAliasState(this._tenantId, this._userId, redirection.id, alias);
                        }
                    }
                }
            }
            return response;
        });
    }
    postMultipart(baseUrl, resourcePath, formData) {
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.VoidClientResponse(400));
        }
        return this._fetchClient.postMultipart(baseUrl, resourcePath, formData);
        ;
    }
    postRecords(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostRecordsPath(resourcePathElems);
        if (pathFields.dialogId === PersistentClient.BRIEFCASE_WORKPACKAGES_DIALOG_ID) {
            let recordSet = persistence_tools_1.PersistenceTools.readRecordSetState(this._tenantId, this._userId, pathFields.dialogId);
            if (!recordSet) {
                recordSet = briefcase_template_1.BriefcaseTemplate.BRIEFCASE_WORKPACKAGES_RECORDSET;
                persistence_tools_1.PersistenceTools.writeRecordSetState(this._tenantId, this._userId, pathFields.dialogId, recordSet);
            }
            return Promise.resolve(new client_1.JsonClientResponse(recordSet, 200));
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return new Promise((resolve, reject) => {
                const recordSet = persistence_tools_1.PersistenceTools.readRecordSetState(this._tenantId, this._userId, pathFields.dialogId);
                if (!recordSet) {
                    resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`Offline record set not found ${pathFields.dialogId}`), 404));
                }
                else {
                    resolve(new client_1.JsonClientResponse(recordSet, 200));
                }
            });
        }
        let response = this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
        return response.then(jcr => {
            if (jcr.statusCode === 200) {
                var recordSet = jcr.value;
                const dialogId = pathFields.dialogId;
                var dialog = persistence_tools_1.PersistenceTools.findRootDialogState(this._tenantId, this._userId, dialogId);
                // If we have persisted the dialog, we also need to persist the dialog records
                if (dialog) {
                    if (this.isWorkPackagesDialog(dialog)) {
                        recordSet = this.patchWorkPackagesRecordSet(dialog, recordSet);
                    }
                    persistence_tools_1.PersistenceTools.writeRecordSetState(this._tenantId, this._userId, dialogId, recordSet);
                }
                return new client_1.JsonClientResponse(recordSet, 200);
            }
            return response;
        });
    }
    postSession(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        this.changeTenantId(resourcePathElems[1]);
        this.changeUserId(jsonBody.userId);
        const briefcaseRecord = this.readBriefcaseRecord();
        if (briefcaseRecord) {
            const onlineProperty = persistence_tools_1.PersistenceTools.findRecordProperty(briefcaseRecord, PersistentClient.ONLINE_PROPERTY_NAME);
            if (onlineProperty && !onlineProperty.value) {
                this.changeClientMode(client_1.ClientMode.OFFLINE);
            }
            else {
                this.changeClientMode(client_1.ClientMode.ONLINE);
            }
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return new Promise((resolve, reject) => {
                const session = persistence_tools_1.PersistenceTools.readSessionState(this._tenantId, this._userId);
                if (!session) {
                    resolve(new client_1.JsonClientResponse(this.createOfflineSessionNotFoundModel(), 404));
                }
                else {
                    this.changeSessionId(session.id);
                    resolve(new client_1.JsonClientResponse(session, 200));
                }
            });
        }
        let response = this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
        return response.then(jcr => {
            if (jcr.statusCode === 200) {
                const session = jcr.value;
                this.changeSessionId(session.id);
                //                PersistenceTools.deleteAllState(this._tenantId, this._userId);
                persistence_tools_1.PersistenceTools.writeSessionState(session);
            }
            return response;
        });
    }
    postWorkbenchAction(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostWorkbenchActionPath(resourcePathElems);
        if (pathFields.actionId === PersistentClient.WORK_PACKAGES_ACTION_ID) {
            return this.postWorkbenchActionWorkPackages(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        else if (pathFields.actionId === PersistentClient.BRIEFCASE_ACTION_ID) {
            return this.postWorkbenchActionBriefcase(baseUrl, resourcePath, resourcePathElems, jsonBody);
        }
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`Workbench action not valid while offline: ${pathFields.actionId}`), 400));
        }
        return this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
    }
    postWorkbenchActionBriefcase(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostWorkbenchActionPath(resourcePathElems);
        const redirection = briefcase_template_1.BriefcaseTemplate.BRIEFCASE_REDIRECTION;
        const navigationId = `${pathFields.actionId}Launcher`;
        const navigation = {
            id: navigationId,
            redirectionId: redirection.id
        };
        persistence_tools_1.PersistenceTools.writeNavigationState(this._tenantId, this._userId, navigation);
        persistence_tools_1.PersistenceTools.writeRedirectionState(this._tenantId, this._userId, redirection);
        persistence_tools_1.PersistenceTools.writeDialogReferringAliasState(this._tenantId, this._userId, redirection.id, 'BriefcaseLauncher');
        persistence_tools_1.PersistenceTools.writeDialogAliasState(this._tenantId, this._userId, redirection.id, 'Briefcase');
        return Promise.resolve(new client_1.JsonClientResponse(redirection, 303));
    }
    postWorkbenchActionWorkPackages(baseUrl, resourcePath, resourcePathElems, jsonBody) {
        const pathFields = this.deconstructPostWorkbenchActionPath(resourcePathElems);
        const navigationId = `${pathFields.actionId}Launcher`;
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return new Promise((resolve, reject) => {
                const navigation = persistence_tools_1.PersistenceTools.readNavigationState(this._tenantId, this._userId, navigationId);
                if (!navigation) {
                    resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`Navigation for offline ${pathFields.actionId} not found`), 404));
                }
                else {
                    const redirection = persistence_tools_1.PersistenceTools.readRedirectionState(this._tenantId, this._userId, navigation.redirectionId);
                    if (!redirection) {
                        resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`Redirection for offline ${pathFields.actionId} not found`), 404));
                    }
                    else {
                        resolve(new client_1.JsonClientResponse(redirection, 303));
                    }
                }
            });
        }
        let response = this._fetchClient.postJson(baseUrl, resourcePath, jsonBody);
        return response.then(jcr => {
            if (jcr.statusCode === 303) {
                if (pathFields.actionId === PersistentClient.WORK_PACKAGES_ACTION_ID) {
                    const redirection = jcr.value;
                    persistence_tools_1.PersistenceTools.deleteAllWorkbenchNavigation(this._tenantId, this._userId, navigationId);
                    const navigation = {
                        id: navigationId,
                        redirectionId: redirection.id
                    };
                    persistence_tools_1.PersistenceTools.writeNavigationState(this._tenantId, this._userId, navigation);
                    persistence_tools_1.PersistenceTools.writeRedirectionState(this._tenantId, this._userId, redirection);
                    persistence_tools_1.PersistenceTools.writeDialogReferringAliasState(this._tenantId, this._userId, redirection.id, 'WorkPackagesLauncher');
                    persistence_tools_1.PersistenceTools.writeDialogAliasState(this._tenantId, this._userId, redirection.id, 'WorkPackages');
                }
            }
            return response;
        });
    }
    putJson(baseUrl, resourcePath, jsonBody) {
        if (this._clientMode === client_1.ClientMode.OFFLINE) {
            return Promise.resolve(new client_1.JsonClientResponse(this.createDialogMessageModel(`PUT action is not valid while offline: ${resourcePath}`), 400));
        }
        return this._fetchClient.putJson(baseUrl, resourcePath, jsonBody);
    }
    /*
    * TODO: Refactor this into persistence-tools as a general method when discriminator and ids can be parameterized
    */
    readBriefcaseRecord() {
        // FIND DIALOG ID
        const briefcaseNavigationId = `${PersistentClient.BRIEFCASE_ACTION_ID}Launcher`;
        const briefcaseNavigation = persistence_tools_1.PersistenceTools.readNavigationState(this._tenantId, this._userId, briefcaseNavigationId);
        if (briefcaseNavigation) {
            const redirectionId = briefcaseNavigation.redirectionId;
            const dialog = persistence_tools_1.PersistenceTools.readDialogState(this._tenantId, this._userId, redirectionId);
            var briefcaseRecordDialogId = null;
            const dialogChildren = dialog.children;
            if (dialogChildren) {
                for (let child of dialogChildren) {
                    // TODO: We need a better discriminator technique for reading record(s) for a
                    // particular dialog.
                    if (child.businessClassName.endsWith('briefcase')) {
                        briefcaseRecordDialogId = child.id;
                        break;
                    }
                }
            }
            // READ RECORD
            if (briefcaseRecordDialogId) {
                return persistence_tools_1.PersistenceTools.readRecordState(this._tenantId, this._userId, briefcaseRecordDialogId);
            }
        }
        return null;
    }
    readBriefcaseWorkpackageRecordSet() {
        return persistence_tools_1.PersistenceTools.readRecordSetState(this._tenantId, this._userId, PersistentClient.BRIEFCASE_WORKPACKAGES_DIALOG_ID);
    }
    setClientMode(clientMode) {
        this.changeClientMode(clientMode);
    }
    writeBriefcaseWorkpackageRecordSet(recordSet) {
        return persistence_tools_1.PersistenceTools.writeRecordSetState(this._tenantId, this._userId, PersistentClient.BRIEFCASE_WORKPACKAGES_DIALOG_ID, recordSet);
    }
    /*
    * TODO: Refactor this into persistence-tools as a general method when discriminator and ids can be parameterized
    */
    writeBriefcaseRecord(briefcaseRecord) {
        // FIND DIALOG ID
        const briefcaseNavigationId = `${PersistentClient.BRIEFCASE_ACTION_ID}Launcher`;
        const briefcaseNavigation = persistence_tools_1.PersistenceTools.readNavigationState(this._tenantId, this._userId, briefcaseNavigationId);
        if (briefcaseNavigation) {
            const redirectionId = briefcaseNavigation.redirectionId;
            const dialog = persistence_tools_1.PersistenceTools.readDialogState(this._tenantId, this._userId, redirectionId);
            var briefcaseRecordDialogId = null;
            const dialogChildren = dialog.children;
            if (dialogChildren) {
                for (let child of dialogChildren) {
                    // TODO: We need a better discriminator technique for reading record(s) for a
                    // particular dialog.
                    if (child.businessClassName.endsWith('briefcase')) {
                        briefcaseRecordDialogId = child.id;
                        break;
                    }
                }
            }
            // WRITE RECORD
            if (briefcaseRecordDialogId) {
                persistence_tools_1.PersistenceTools.writeRecordState(this._tenantId, this._userId, briefcaseRecordDialogId, briefcaseRecord);
            }
        }
    }
}
PersistentClient.ADD_TO_BRIEFCASE_ACTION_ID = 'alias_AddToBriefcase';
PersistentClient.BRIEFCASE_ACTION_ID = 'Briefcase';
PersistentClient.BRIEFCASE_DIALOG_ID = 'a1';
PersistentClient.BRIEFCASE_RECORD_DIALOG_ID = 'a11';
PersistentClient.BRIEFCASE_WORKPACKAGES_DIALOG_ID = 'a12';
PersistentClient.ENTER_OFFLINE_MODE_ACTION_ID = 'alias_EnterOfflineMode';
PersistentClient.EXIT_OFFLINE_MODE_ACTION_ID = 'alias_ExitOfflineMode';
PersistentClient.OPEN_ACTION_ID = 'open';
PersistentClient.OPEN_DOCUMENT_ACTION_ID = 'dynamic_AAABACcaAAAAACqd';
PersistentClient.ONLINE_PROPERTY_NAME = 'online';
PersistentClient.REFERRING_WORKBENCH_MODEL_TYPE = 'hxgn.api.dialog.ReferringWorkbench';
PersistentClient.REFERRING_DIALOG_MODEL_TYPE = 'hxgn.api.dialog.ReferringDialog';
PersistentClient.SDA_WORKBENCH_ID = 'AAABACffAAAAACe2';
PersistentClient.WORK_PACKAGES_ACTION_ID = 'WorkPackages';
exports.PersistentClient = PersistentClient;
