"use strict";
/**
 * Created by rburson on 8/23/17.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("./util");
const client_1 = require("./client");
class FetchClient {
    constructor(clientMode = client_1.ClientMode.ONLINE) {
        this._lastActivity = new Date();
        this._clientMode = clientMode;
    }
    getBlob(baseUrl, resourcePath) {
        const url = resourcePath ? `${baseUrl}/${resourcePath}` : baseUrl;
        return this.processRequest(url, 'GET').then((response) => {
            return response.blob().then(blob => new client_1.BlobClientResponse(blob, response.status));
        });
    }
    getText(baseUrl, resourcePath) {
        const url = resourcePath ? `${baseUrl}/${resourcePath}` : baseUrl;
        return this.processRequest(url, 'GET').then((response) => {
            return response.text().then(text => new client_1.TextClientResponse(text, response.status));
        });
    }
    postMultipart(baseUrl, resourcePath, formData) {
        const url = resourcePath ? `${baseUrl}/${resourcePath}` : baseUrl;
        return this.processRequest(url, 'POST', formData).then((response) => {
            return new client_1.VoidClientResponse(response.status);
        });
    }
    getJson(baseUrl, resourcePath, queryParams) {
        const headers = { 'Accept': 'application/json' };
        const queryString = this.encodeQueryParams(queryParams);
        const url = resourcePath ? `${baseUrl}/${resourcePath}${queryString}` : `${baseUrl}${queryString}`;
        return this.processRequest(url, 'GET', null, headers).then((response) => {
            return this.assertJsonContentType(response.headers.get('content-type')).then(() => {
                return response.json().then(json => new client_1.JsonClientResponse(json, response.status));
            });
        });
    }
    get lastActivity() {
        return this._lastActivity;
    }
    postJson(baseUrl, resourcePath, jsonBody) {
        const headers = { 'Accept': 'application/json', 'Content-Type': 'application/json;charset=UTF-8' };
        const body = jsonBody && JSON.stringify(jsonBody);
        const url = resourcePath ? `${baseUrl}/${resourcePath}` : baseUrl;
        return this.processRequest(url, 'POST', body, headers).then((response) => {
            return this.assertJsonContentType(response.headers.get('content-type')).then(() => {
                return response.json().then(json => new client_1.JsonClientResponse(json, response.status));
            });
        });
    }
    putJson(baseUrl, resourcePath, jsonBody) {
        const headers = { 'Accept': 'application/json', 'Content-Type': 'application/json;charset=UTF-8' };
        const body = jsonBody && JSON.stringify(jsonBody);
        const url = resourcePath ? `${baseUrl}/${resourcePath}` : baseUrl;
        return this.processRequest(url, 'PUT', body, headers).then((response) => {
            return this.assertJsonContentType(response.headers.get('content-type')).then(() => {
                return response.json().then(json => new client_1.JsonClientResponse(json, response.status));
            });
        });
    }
    deleteJson(baseUrl, resourcePath) {
        const headers = { 'Accept': 'application/json' };
        const url = resourcePath ? `${baseUrl}/${resourcePath}` : baseUrl;
        return this.processRequest(url, 'DELETE', null, headers).then((response) => {
            return this.assertJsonContentType(response.headers.get('content-type')).then(() => {
                return response.json().then(json => new client_1.JsonClientResponse(json, response.status));
            });
        });
    }
    setClientMode(clientMode) {
        this._clientMode = clientMode;
    }
    assertJsonContentType(contentType) {
        return new Promise((resolve, reject) => {
            if (this.isJsonContentType(contentType)) {
                resolve();
            }
            else {
                reject(new Error(`Expected 'application/json', got ${contentType}`));
            }
        });
    }
    encodeQueryParams(queryParams) {
        let result = '';
        if (queryParams) {
            for (const name in queryParams) {
                result += `${encodeURIComponent(name)}=${encodeURIComponent(queryParams[name])}&`;
            }
        }
        return result.length > 0 ? `?${result.slice(0, -1)}` : result;
    }
    isJsonContentType(contentType) {
        return contentType && contentType.includes('application/json');
    }
    processRequest(url, method, body, headers) {
        return new Promise((resolve, reject) => {
            const requestHeaders = new Headers(headers);
            requestHeaders.append('Accept', 'gzip');
            const init = { method: method, mode: 'cors' };
            if (body)
                init.body = body;
            if (headers)
                init.headers = new Headers(headers);
            if (!['GET', 'POST', 'PUT', 'DELETE'].some(v => method === v)) {
                reject(new Error(`FetchClient::processRequest: Unsupported method: ${method}`));
            }
            else {
                util_1.Log.debug(`Fetch request: ${method} ${url} [body]:${body ? body : 'none'}`);
                fetch(url, init)
                    .then(response => { this._lastActivity = new Date(); resolve(response); })
                    .catch(error => { this._lastActivity = new Date(); reject(error); });
            }
        });
    }
}
exports.FetchClient = FetchClient;
