/**
 * Created by rburson on 8/29/17.
 */
import { Client, ClientMode, JsonClientResponse } from "./client";
import { ClientType, DialogRedirection, Login, Menu, Redirection, Session, Workbench, WorkbenchAction, View, ViewMode, ViewDescriptor, Record, ActionParameters, RecordSet, QueryParameters, Dialog, EditorDialog } from "./models";
import 'moment/locale/zh-cn';
import 'moment/locale/ru';
import 'moment/locale/fr';
import 'moment/locale/es';
import 'moment/locale/en-gb';
import 'moment/locale/de';
import 'moment/locale/pt';
import 'moment/locale/pt-br';
import 'moment/locale/en-ca';
import 'moment/locale/it';
import 'moment/locale/ja';
/**
 * Top-level entry point into the Catavolt API
 */
export declare class CatavoltApi {
    private static _singleton;
    private static ONE_HOUR_IN_MILLIS;
    private static SERVER_URL;
    private static SERVER_VERSION;
    dataLastChangedTime: Date;
    private _clientMode;
    private _dialogApi;
    private _session;
    private _devicePropsDynamic;
    private _devicePropsStatic;
    /**
     * Get the default session time
     * @returns {number}
     */
    static readonly defaultTTLInMillis: number;
    /**
     * Get the singleton instance of the CatavoltApi
     * @returns {CatavoltApi}
     */
    static readonly singleton: CatavoltApi;
    /**
     * Construct an CatavoltApi
     * This should not be called directly, instead use the 'singleton' method
     * @private
     */
    private constructor();
    /**
     * Add or replace a dynamic device property (func)
     * @param propName
     * @param propFn
     */
    addDynamicDeviceProp(propName: string, propFn: () => string): void;
    /**
     * Add or replace a static device property
     *
     * @param propName
     * @param propValue
     */
    addStaticDeviceProp(propName: string, propValue: string): void;
    /**
     * Get the json representation of this client's locale.  The server pulls this value from the agent string
     * and returns it to the client.
     * @returns {string}
     */
    readonly browserLocaleJson: string;
    changePasswordAndLogin(tenantId: string, clientType: ClientType, userId: string, existingPassword: string, newPassword: string): Promise<Session | Redirection>;
    /**
     * Get the number of millis that the client will remain active between calls
     * to the server.
     * @returns {number}
     */
    readonly clientTimeoutMillis: number;
    /**
     * Get the currency symbol override if defined from the server.
     * @returns {string}
     */
    readonly currencySymbol: string;
    /**
     * Get the device props
     * @returns {{[p: string]: string}}
     */
    readonly deviceProps: {
        [index: string]: string;
    };
    /**
     * Get the DialogApi instance
     * @returns {DialogApi}
     */
    readonly dialogApi: DialogApi;
    /**
     * Get a Workbench by workbenchId
     * @param workbenchId
     * @returns {Promise<Workbench>}
     */
    getWorkbench(workbenchId: string): Promise<Workbench>;
    /**
     * Get the list of Workbenches
     *
     * @returns {Array<Workbench>}
     */
    getWorkbenches(): Promise<Array<Workbench>>;
    /**
     * Initialize a dialog service implementation for use by this CatavoltApi
     *
     * @param serverVersion
     * @param serverUrl
     */
    initDialogApi(serverUrl: string, serverVersion?: string): void;
    /**
     * Initialize an offline dialog service
     *
     * @param serverVersion
     * @param serverUrl
     */
    initPersistentApi(serverUrl: string, serverVersion: string): void;
    isOfflineMode(): boolean;
    isPersistentClient(): boolean;
    /**
     * Check for the availability of the given featureSet
     * @see FeatureSet
     * @param featureSet
     * @returns {boolean}
     */
    isFeatureSetAvailable(featureSet: FeatureSet): boolean;
    /**
     * Checked logged in status
     * @returns {boolean}
     */
    readonly isLoggedIn: boolean;
    /**
     * Log in and retrieve the Session
     *
     * @param tenantId
     * @param clientType
     * @param userId
     * @param password
     * @param serverUrl
     * @param serverVersion
     *
     * @returns {Promise<Session | Redirection>}
     */
    login(tenantId: string, clientType: ClientType, userId: string, password: string): Promise<Session | Redirection>;
    /**
     * Logout and destroy the session
     * @returns {{sessionId:string}}
     */
    logout(): Promise<{
        sessionId: string;
    }>;
    openDialog(redirection: DialogRedirection): Promise<Dialog>;
    toDialogOrRedirection(resultPr: Promise<{}>): Promise<Dialog | Redirection>;
    getRedirection(redirectionId: string): Promise<Redirection>;
    /**
     * Open a {@link WorkbenchAction}
     * @param workbenchAction
     * @returns {Promise<{actionId:string} | Redirection>}
     */
    performWorkbenchAction(workbenchAction: WorkbenchAction): Promise<{
        actionId: string;
    } | Redirection>;
    /**
     * Open a {@link WorkbenchWorkbenchAction}
     * @param workbenchId
     * @param workbenchActionId
     * @returns {Promise<{actionId:string} | Redirection>}
     */
    performWorkbenchActionForId(workbenchId: string, workbenchActionId: string): Promise<{
        actionId: string;
    } | Redirection>;
    /**
     * Refresh the CatavoltApi
     *
     * @returns {Promise<Session>}
     */
    refreshSession(tenantId: string, sessionId: string): Promise<Session>;
    /**
     * Time remaining before this session is expired by the server
     * @returns {number}
     */
    readonly remainingSessionTime: number;
    /**
     * Get the Session
     * @returns {Session}
     */
    readonly session: Session;
    /**
     * Return whether or not the session has expired
     * @returns {boolean}
     */
    readonly sessionHasExpired: boolean;
    setPersistentClient(): void;
    setOnlineClient(): void;
    setOfflineMode(): void;
    setOnlineMode(): void;
}
export interface DialogApi {
    createSession(tenantId: string, login: Login): Promise<Session | Redirection>;
    getSession(tenantId: string, sessionId: string): Promise<Session>;
    deleteSession(tenantId: string, sessionId: string): Promise<{
        sessionId: string;
    }>;
    getWorkbenches(tenantId: string, sessionId: string): Promise<Array<Workbench>>;
    getWorkbench(tenantId: string, sessionId: string, workbenchId: string): Promise<Workbench>;
    getDialog(tenantId: string, sessionId: string, dialogId: string): Promise<Dialog>;
    deleteDialog(tenantId: string, sessionId: string, dialogId: string): Promise<{
        dialogId: string;
    }>;
    getActions(tenantId: string, sessionId: string, dialogId: string): Promise<Array<Menu>>;
    performAction(tenantId: string, sessionId: string, dialogId: string, actionId: string, actionParameters: ActionParameters): Promise<{
        actionId: string;
    } | Redirection>;
    getWorkbenchActions(tenantId: string, sessionId: string, workbenchId: string): Promise<Array<WorkbenchAction>>;
    performWorkbenchAction(tenantId: string, sessionId: string, workbenchId: string, actionId: string): Promise<{
        actionId: string;
    } | Redirection>;
    getRedirection(tenantId: string, sessionId: string, redirectionId: string): Promise<Redirection>;
    getRecord(tenantId: string, sessionId: string, dialogId: string): Promise<Record>;
    putRecord(tenantId: string, sessionId: string, dialogId: string, record: Record): Promise<Record | Redirection>;
    getRecords(tenantId: string, sessionId: string, dialogId: string, queryParams: QueryParameters): Promise<RecordSet>;
    getMode(tenantId: string, sessionId: string, dialogId: string): Promise<ViewMode>;
    changeMode(tenantId: string, sessionId: string, dialogId: string, mode: ViewMode): Promise<EditorDialog>;
    getView(tenantId: string, sessionId: string, dialogId: string): Promise<View>;
    changeView(tenantId: string, sessionId: string, dialogId: string, viewId: string): Promise<Dialog>;
    getViews(tenantId: string, sessionId: string, dialogId: string): Promise<Array<ViewDescriptor>>;
    setClientMode(clientMode: ClientMode): void;
    lastServiceActivity: Date;
}
export declare class DialogService implements DialogApi {
    readonly client: Client;
    readonly apiVersion: string;
    private static SERVER;
    readonly baseUrl: string;
    constructor(client: Client, serverUrl?: string, apiVersion?: string);
    createSession(tenantId: string, login: Login): Promise<Session | Redirection>;
    getSession(tenantId: string, sessionId: string): Promise<Session>;
    deleteSession(tenantId: string, sessionId: string): Promise<{
        sessionId: string;
    }>;
    getWorkbenches(tenantId: string, sessionId: string): Promise<Array<Workbench>>;
    getWorkbench(tenantId: string, sessionId: string, workbenchId: string): Promise<Workbench>;
    getRedirection(tenantId: string, sessionId: string, redirectionId: string): Promise<Redirection>;
    getDialog(tenantId: string, sessionId: string, dialogId: string): Promise<Dialog>;
    deleteDialog(tenantId: string, sessionId: string, dialogId: string): Promise<{
        dialogId: string;
    }>;
    getActions(tenantId: string, sessionId: string, dialogId: string): Promise<Array<Menu>>;
    performAction(tenantId: string, sessionId: string, dialogId: string, actionId: string, actionParameters: ActionParameters): Promise<{
        actionId: string;
    } | Redirection>;
    getWorkbenchActions(tenantId: string, sessionId: string, workbenchId: string): Promise<Array<WorkbenchAction>>;
    performWorkbenchAction(tenantId: string, sessionId: string, workbenchId: string, actionId: string): Promise<{
        actionId: string;
    } | Redirection>;
    getRecord(tenantId: string, sessionId: string, dialogId: string): Promise<Record>;
    putRecord(tenantId: string, sessionId: string, dialogId: string, record: Record): Promise<Record | Redirection>;
    getRecords(tenantId: string, sessionId: string, dialogId: string, queryParams: QueryParameters): Promise<RecordSet>;
    getMode(tenantId: string, sessionId: string, dialogId: string): Promise<ViewMode>;
    changeMode(tenantId: string, sessionId: string, dialogId: string, mode: ViewMode): Promise<EditorDialog>;
    getView(tenantId: string, sessionId: string, dialogId: string): Promise<View>;
    changeView(tenantId: string, sessionId: string, dialogId: string, viewId: string): Promise<Dialog>;
    getViews(tenantId: string, sessionId: string, dialogId: string): Promise<Array<ViewDescriptor>>;
    readonly lastServiceActivity: Date;
    setClientMode(clientMode: ClientMode): void;
    private get(path, queryParams?);
    private post<T>(path, body?);
    private d3lete(path);
    private put<T>(path, body?);
}
export interface DialogApiResponse<T> {
    responseValue(): Promise<T>;
    responseValueOrRedirect(): Promise<T | Redirection>;
    assertNoError(): Promise<void>;
}
export declare class DialogServiceResponse<T> implements DialogApiResponse<T> {
    private readonly clientResponse;
    constructor(clientResponse: JsonClientResponse);
    responseValue(): Promise<T>;
    responseValueOrRedirect(): Promise<T | Redirection>;
    assertNoError(): Promise<void>;
    readonly hasValue: boolean;
    readonly hasRedirection: boolean;
    readonly hasError: boolean;
    private fullfillJsonToModel<T>(clientResponse, resolve, reject);
}
/**
 * Available Features
 */
export declare type FeatureSet = "View_Support" | "Unified_Search";
export declare const Catavolt: CatavoltApi;
