/**
 * Created by rburson on 10/19/17.
 */
import { StringDictionary } from './util';
import { CatavoltApi } from "./dialog";
/** ************** Base classes have to be defined first i.e. Order matters ********************/
export declare abstract class BinaryRef {
    private _settings;
    constructor(_settings: StringDictionary);
    readonly settings: StringDictionary;
}
export declare abstract class CellValue {
    readonly style: string;
    static STYLE_HEADING1: string;
    static STYLE_HEADING2: string;
    static STYLE_HEADING3: string;
    static STYLE_HEADING4: string;
    static STYLE_INLINE_MEDIA: string;
    static STYLE_INLINE_MEDIA2: string;
    readonly type: string;
    constructor(style: string);
    readonly isHeading1Style: boolean;
    readonly isHeading2Style: boolean;
    readonly isHeading3Style: boolean;
    readonly isHeading4Style: boolean;
    readonly isInlineMediaStyle: boolean;
}
/**
 * A View represents a Catavolt 'Pane' definition.  A Pane can be thought of as a 'panel' or UI component
 * that is responsible for displaying a data record or records. The Pane describes 'how' and 'where' the data will be
 * displayed, as well as surrounding 'meta' data (i.e. the Pane title, the Pane's menus).  The Pane itself does not contain
 * the record or records to be displayed, but may be combined with a {@link Record}(s) to display the data.
 */
export declare abstract class View {
    readonly alias: string;
    readonly id: string;
    readonly name: string;
    readonly menu: Menu;
    readonly title: string;
    readonly type: ViewType;
    /**
     * Find the title for this Pane
     * @returns {string}
     */
    findTitle(): string;
    /**
     * Find a menu def on this View with the given actionId
     * @param actionId
     * @returns {Menu}
     */
    findMenuAt(actionId: string): Menu;
}
/** ************************** Subclasses *******************************************************/
export interface ActionParameters {
    readonly pendingWrites?: Record;
    readonly targets?: Array<string>;
    readonly type: string;
}
export interface AppWindow {
    readonly initialAction: WorkbenchAction;
    readonly notificationsAction: WorkbenchAction;
    readonly windowHeight: number;
    readonly windowWidth: number;
    readonly windowTitle: string;
    readonly workbenches: ReadonlyArray<Workbench>;
}
/**
 * Defines how to present a business-value in a UI component
 */
export declare class AttributeCellValue extends CellValue {
    readonly propertyName: string;
    readonly entryMethod: AttributeCellValueEntryMethod;
    readonly hint: string;
    readonly tooltip: string;
    readonly mask: string;
    readonly autoFillCapable: boolean;
    readonly actions: Array<Menu>;
    constructor(propertyName: string, entryMethod: AttributeCellValueEntryMethod, hint: string, tooltip: string, mask: string, autoFillCapable: boolean, actions: Array<Menu>, style: string);
    readonly isComboBoxEntryMethod: boolean;
    readonly isDropDownEntryMethod: boolean;
    readonly isIconEntryMethod: boolean;
    readonly isTextFieldEntryMethod: boolean;
}
/**
 * A purely declarative type. This object has no additional properties.
 */
export declare class BarcodeScan extends View {
}
/**
 * ************* Binary Support ********************
 */
/**
 * Represents a binary value
 */
export interface Binary {
    /**
     * Return a url resprenting this binary value
     */
    toUrl(): string;
}
/**
 * Represents a base64 encoded binary
 */
export declare class EncodedBinary implements Binary {
    private _data;
    private _mimeType;
    constructor(_data: string, _mimeType?: string);
    /**
     * Get the base64 encoded data
     * @returns {string}
     */
    readonly data: string;
    /**
     * Get the mime-type
     * @returns {string|string}
     */
    readonly mimeType: string;
    /**
     * Returns a 'data url' representation of this binary, including the encoded data
     * @returns {string}
     */
    toUrl(): string;
}
/**
 * Represents a remote binary
 */
export declare class UrlBinary implements Binary {
    private _url;
    constructor(_url: string);
    readonly url: string;
    /**
     * Returns a url that 'points to' the binary data
     * @returns {string}
     */
    toUrl(): string;
}
export declare class Attachment {
    name: string;
    attachmentData: any;
    constructor(name: string, attachmentData: any);
}
/**
 * An abstract visual Calendar
 */
export declare class Calendar extends View {
    readonly descriptionPropertyName: string;
    readonly initialStyle: string;
    readonly endDatePropertyName: string;
    readonly endTimePropertyName: string;
    readonly occurDatePropertyName: string;
    readonly occurTimePropertyName: string;
    readonly startDatePropertyName: string;
    readonly startTimePropertyName: string;
}
export interface Cell {
    values: Array<CellValue>;
}
export interface Column {
    readonly propertyName: string;
    readonly heading: string;
}
export declare class CodeRef {
    readonly code: string;
    readonly description: string;
    readonly type: string;
    constructor(code: string, description: string, type: string);
    toString(): string;
}
/**
 * A abstract definition for small visual area populated with labels and values. Labels are typically presented as simple text phrases.
 * Values are usually presented inside a UI component, such as a TextField or ComboBox.
 */
export declare class Details extends View {
    readonly cancelButtonText: string;
    readonly commitButtonText: string;
    readonly editable: boolean;
    readonly focusPropertyName: string;
    readonly gmlMarkup: string;
    readonly rows: Array<Array<Cell>>;
}
export declare class ReferringObject {
    readonly type: string;
    readonly actionId: string;
    isDialogReferrer(): boolean;
    isWorkbenchReferrer(): boolean;
}
export declare class ReferringDialog extends ReferringObject {
    readonly dialogId: string;
    readonly dialogMode: DialogMode;
}
export declare class ReferringWorkbench extends ReferringObject {
    readonly workbenchId: string;
}
export interface DialogMessage {
    /**
     * A short language-independent identifier
     */
    readonly code: string;
    readonly messageType: DialogMessageMessageType;
    /**
     * A human-readable informative description. If a code is provided, then this message explains the meaning of the code.
     */
    readonly message: string;
    /**
     * An object typically provided to help programmers diagnose an error.
     * For example, a cause can be the name of a host programming exception.
     */
    readonly cause: any;
    /**
     * This property is provided when the message pertains to one or more properties in a user interface view.
     */
    readonly propertyNames: Array<string>;
    /**
     * If this message is a generalization or aggregation, then children messages can be used to explain the individual facets.
     */
    readonly children: Array<DialogMessage>;
    /**
     * If the case of a host programming error, this property contains a stack trace of the host programming language.
     */
    readonly stackTrace: string;
    readonly type: string;
}
export interface DialogRedirection extends Redirection {
    readonly dialogDescription: string;
    readonly dialogId: string;
}
export declare class DialogException {
    iconName: string;
    message: string;
    name: string;
    stackTrace: string;
    title: string;
    cause: DialogException;
    userMessages: Array<UserMessage>;
    constructor(iconName?: string, message?: string, name?: string, stackTrace?: string, title?: string, cause?: DialogException, userMessages?: Array<UserMessage>);
}
export interface Annotation {
    readonly name: string;
    readonly value: string;
    readonly type: string;
}
export declare class DataAnnotation implements Annotation {
    readonly name: string;
    readonly value: string;
    readonly type: string;
    private static BOLD_TEXT;
    private static BACKGROUND_COLOR;
    private static FOREGROUND_COLOR;
    private static IMAGE_NAME;
    private static IMAGE_PLACEMENT;
    private static ITALIC_TEXT;
    private static OVERRIDE_TEXT;
    private static TIP_TEXT;
    private static UNDERLINE;
    private static TRUE_VALUE;
    private static PLACEMENT_CENTER;
    private static PLACEMENT_LEFT;
    private static PLACEMENT_RIGHT;
    private static PLACEMENT_UNDER;
    private static PLACEMENT_STRETCH_UNDER;
    static backgroundColor(annotations: Array<DataAnnotation>): string;
    static foregroundColor(annotations: Array<DataAnnotation>): string;
    static imageName(annotations: Array<DataAnnotation>): string;
    static imagePlacement(annotations: Array<DataAnnotation>): string;
    static isBoldText(annotations: Array<DataAnnotation>): boolean;
    static isItalicText(annotations: Array<DataAnnotation>): boolean;
    static isPlacementCenter(annotations: Array<DataAnnotation>): boolean;
    static isPlacementLeft(annotations: Array<DataAnnotation>): boolean;
    static isPlacementRight(annotations: Array<DataAnnotation>): boolean;
    static isPlacementStretchUnder(annotations: Array<DataAnnotation>): boolean;
    static isPlacementUnder(annotations: Array<DataAnnotation>): boolean;
    static isUnderlineText(annotations: Array<DataAnnotation>): boolean;
    static overrideText(annotations: Array<DataAnnotation>): string;
    static tipText(annotations: Array<DataAnnotation>): string;
    constructor(name: string, value: string, type: string);
    readonly backgroundColor: string;
    readonly foregroundColor: string;
    equals(dataAnno: Annotation): boolean;
    readonly isBackgroundColor: boolean;
    readonly isBoldText: boolean;
    readonly isForegroundColor: boolean;
    readonly isImageName: boolean;
    readonly isImagePlacement: boolean;
    readonly isItalicText: boolean;
    readonly isOverrideText: boolean;
    readonly isPlacementCenter: boolean;
    readonly isPlacementLeft: boolean;
    readonly isPlacementRight: boolean;
    readonly isPlacementStretchUnder: boolean;
    readonly isPlacementUnder: boolean;
    readonly isTipText: boolean;
    readonly isUnderlineText: boolean;
    toJSON(): Annotation;
}
/**
 * Utility for working with Records
 */
export declare class RecordUtil {
    static newRecord(id: string, properties: Array<Property>, annotations: Array<DataAnnotation>, type: string): Record;
    static isRecord(o: any): boolean;
}
/**
 * An {@link Record} that manages two copies internally, a before and after, for 'undo' and comparison purposes.
 * An Record Represents a 'Record' or set of {@link Prop} (names and values).
 * An Record may also have {@link Annotation}s (style annotations) that apply to the whole 'record'
 */
export declare class RecordBuffer implements Record {
    private _before;
    private _after;
    static createRecordBuffer(id: string, before: Array<Property>, after: Array<Property>, annotations: Array<DataAnnotation>, type: string): RecordBuffer;
    constructor(_before: Record, _after?: Record);
    readonly after: Record;
    readonly annotations: Array<DataAnnotation>;
    annotationsAtName(propName: string): Array<DataAnnotation>;
    afterEffects(afterAnother?: Record): Record;
    readonly backgroundColor: string;
    backgroundColorFor(propName: string): string;
    readonly before: Record;
    readonly foregroundColor: string;
    foregroundColorFor(propName: string): string;
    readonly imageName: string;
    imageNameFor(propName: string): string;
    readonly imagePlacement: string;
    imagePlacementFor(propName: string): string;
    readonly isBoldText: boolean;
    isBoldTextFor(propName: string): boolean;
    isChanged(name: string): boolean;
    readonly isItalicText: boolean;
    isItalicTextFor(propName: string): boolean;
    readonly isPlacementCenter: boolean;
    isPlacementCenterFor(propName: string): boolean;
    readonly isPlacementLeft: boolean;
    isPlacementLeftFor(propName: string): boolean;
    readonly isPlacementRight: boolean;
    isPlacementRightFor(propName: string): boolean;
    readonly isPlacementStretchUnder: boolean;
    isPlacementStretchUnderFor(propName: string): boolean;
    readonly isPlacementUnder: boolean;
    isPlacementUnderFor(propName: string): boolean;
    readonly isUnderline: boolean;
    isUnderlineFor(propName: string): boolean;
    readonly id: string;
    readonly overrideText: string;
    overrideTextFor(propName: string): string;
    propAtIndex(index: number): Property;
    propAtName(propName: string): Property;
    readonly propCount: number;
    readonly propNames: Array<string>;
    readonly properties: Array<Property>;
    readonly propValues: Array<any>;
    readonly type: string;
    setValue(name: string, value: any): void;
    readonly tipText: string;
    tipTextFor(propName: string): string;
    toRecord(): Record;
    toJSON(): any;
    valueAtName(propName: string): any;
}
/**
 * *********************************
 */
/**
 * The implementation of {@link Record}.
 * Represents a 'Record' or set of {@link Prop} (names and values).
 * An Record may also have {@link Annotation}s (style annotations) that apply to the whole 'record'
 */
export declare class RecordImpl implements Record {
    readonly id: string;
    readonly properties: Array<Property>;
    readonly annotations: Array<DataAnnotation>;
    readonly type: string;
    constructor(id: string, properties: Array<Property>, annotations: Array<DataAnnotation>, type: string);
    annotationsAtName(propName: string): Array<DataAnnotation>;
    afterEffects(after: Record): Record;
    readonly backgroundColor: string;
    backgroundColorFor(propName: string): string;
    readonly foregroundColor: string;
    foregroundColorFor(propName: string): string;
    readonly imageName: string;
    imageNameFor(propName: string): string;
    readonly imagePlacement: string;
    imagePlacementFor(propName: string): string;
    readonly isBoldText: boolean;
    isBoldTextFor(propName: string): boolean;
    readonly isItalicText: boolean;
    isItalicTextFor(propName: string): boolean;
    readonly isPlacementCenter: boolean;
    isPlacementCenterFor(propName: string): boolean;
    readonly isPlacementLeft: boolean;
    isPlacementLeftFor(propName: string): boolean;
    readonly isPlacementRight: boolean;
    isPlacementRightFor(propName: string): boolean;
    readonly isPlacementStretchUnder: boolean;
    isPlacementStretchUnderFor(propName: string): boolean;
    readonly isPlacementUnder: boolean;
    isPlacementUnderFor(propName: string): boolean;
    readonly isUnderline: boolean;
    isUnderlineFor(propName: string): boolean;
    readonly overrideText: string;
    overrideTextFor(propName: string): string;
    propAtIndex(index: number): Property;
    propAtName(propName: string): Property;
    readonly propCount: number;
    readonly propNames: Array<string>;
    readonly propValues: Array<any>;
    readonly tipText: string;
    tipTextFor(propName: string): string;
    toJSON(): {
        id: string;
        properties: Property[];
        type: TypeNames;
    };
    toRecord(): Record;
    valueAtName(propName: string): any;
}
/**
 * *********************************
 */
export declare class ErrorMessage extends View {
    readonly exception: DialogException;
}
export interface Filter {
    readonly not: boolean;
    readonly operand1: any;
    readonly operator: FilterOperator;
    readonly operand2: any;
}
/**
 * A purely declarative type. This object has no additional properties.
 */
export declare class ForcedLineCellValue extends CellValue {
    constructor(style?: string);
}
/**
 * A composition of View objects that, together, comprise a UI form.
 */
export declare class Form extends View {
    readonly borderStyle: string;
    readonly formStyle: string;
    readonly formLayout: string;
    readonly isCompositeForm: boolean;
    readonly isFlowingLayout: boolean;
    readonly isFlowingTopDownLayout: boolean;
    readonly isFourBoxSquareLayout: boolean;
    readonly isHorizontalLayout: boolean;
    readonly isOptionsFormLayout: boolean;
    readonly isTabsLayout: boolean;
    readonly isThreeBoxOneLeftLayout: boolean;
    readonly isThreeBoxOneOverLayout: boolean;
    readonly isThreeBoxOneRightLayout: boolean;
    readonly isThreeBoxOneUnderLayout: boolean;
    readonly isTopDownLayout: boolean;
    readonly isTwoVerticalLayout: boolean;
}
/**
 * A purely declarative type. This object has no additional properties.
 */
export declare class GpsReading extends View {
}
export declare class GpsReadingProperty {
    readonly accuracy: number;
    readonly latitude: number;
    readonly longitude: number;
    readonly source: string;
    readonly type: string;
    constructor(accuracy: number, latitude: number, longitude: number, source: string, type: string);
}
/**
 * A purely declarative type. This object has no additional properties.
 */
export declare class MapLocation extends View {
}
export declare class MapLocationProperty {
    readonly latitude: number;
    readonly longitude: number;
    readonly type: string;
    constructor(latitude: number, longitude: number, type: string);
}
/**
 * A view describing how to display a collection of data as a line graph, pie chart, bar chart, etc.
 */
export declare class Graph extends View {
    static GRAPH_TYPE_CARTESIAN: string;
    static GRAPH_TYPE_PIE: string;
    static PLOT_TYPE_BAR: string;
    static PLOT_TYPE_BUBBLE: string;
    static PLOT_TYPE_LINE: string;
    static PLOT_TYPE_SCATTER: string;
    static PLOT_TYPE_STACKED: string;
    readonly graphType: string;
    readonly displayQuadrantLines: boolean;
    readonly identityDataPoint: GraphDataPoint;
    readonly groupingDataPoint: GraphDataPoint;
    readonly dataPoints: Array<GraphDataPoint>;
    readonly filterDataPoints: Array<GraphDataPoint>;
    readonly sampleModel: string;
    readonly xAxisLabel: string;
    readonly xAxisRangeFrom: number;
    readonly xAxisRangeTo: number;
    readonly yAxisLabel: string;
    readonly yAxisRangeFrom: number;
    readonly yAxisRangeTo: number;
}
export interface GraphDataPoint {
    readonly propertyName: string;
    readonly legendKey: string;
    readonly plotType: string;
    readonly type: string;
    readonly bubbleRadiusName: string;
    readonly bubbleRadiusType: string;
    readonly seriesColor: string;
    readonly xAxisName: string;
    readonly xAxisType: string;
}
export declare class ImagePicker extends View {
}
export declare class InlineBinaryRef extends BinaryRef {
    private _inlineData;
    constructor(_inlineData: string, settings: StringDictionary);
    readonly inlineData: string;
    toString(): string;
}
/**
 * A text description typically preceeding a UI component as a prompt
 */
export declare class LabelCellValue extends CellValue {
    readonly value: string;
    constructor(style: any, value: string);
}
export interface Login {
    readonly userId: string;
    readonly password: string;
    readonly clientType: ClientType;
    readonly deviceProperties: StringDictionary;
    readonly type: string;
}
/**
 * Columns, filter and sorts for a UI list component.
 */
export declare class List extends View {
    readonly style: string;
    readonly columnStyle: string;
    readonly gmlMarkup: string;
    readonly fixedColumnCount: number;
    readonly columns: Array<Column>;
    readonly filter: Array<Filter>;
    readonly sort: Array<Sort>;
    readonly isDefaultStyle: boolean;
    readonly isDetailsFormStyle: boolean;
    readonly isFormStyle: boolean;
    readonly isTabularStyle: boolean;
    readonly columnHeadings: Array<string>;
    rowValues(record: Record): Array<any>;
}
export declare class Map extends View {
    readonly cityPropertyName: string;
    readonly descriptionPropertyName: string;
    readonly latitudePropertyName: string;
    readonly longitudePropertyName: string;
    readonly postalCodePropertyName: string;
    readonly statePropertyName: string;
    readonly streetPropertyName: string;
}
export declare class Menu {
    readonly children: Array<Menu>;
    /**
     * A special value understood by the UI, such as 'refresh'
     */
    readonly actionId: string;
    readonly directive: string;
    readonly iconUrl: string;
    readonly id: string;
    readonly label: string;
    readonly visible: boolean;
    /**
     * The menu is allowed (active) for these modes
     */
    readonly modes: Array<string>;
    readonly type: string;
    static findSubMenu(md: Menu, matcher: (menu: Menu) => boolean): Menu;
    findAtActionId(actionId: string): Menu;
    findContextMenu(): Menu;
    readonly isPresaveDirective: boolean;
    readonly isRead: boolean;
    readonly isSeparator: boolean;
    readonly isWrite: boolean;
}
/**
 * An empty or uninitialized {@link Record}.
 * Represents a 'Record' or set of {@link Prop} (names and values).
 * An Record may also have {@link Annotation}s (style annotations) that apply to the whole 'record'
 */
export declare class NullRecord implements Record {
    static singleton: NullRecord;
    constructor();
    readonly annotations: Array<DataAnnotation>;
    annotationsAtName(propName: string): Array<DataAnnotation>;
    afterEffects(after: Record): Record;
    readonly backgroundColor: string;
    backgroundColorFor(propName: string): string;
    readonly foregroundColor: string;
    foregroundColorFor(propName: string): string;
    readonly id: string;
    readonly imageName: string;
    imageNameFor(propName: string): string;
    readonly imagePlacement: string;
    imagePlacementFor(propName: string): string;
    readonly isBoldText: boolean;
    isBoldTextFor(propName: string): boolean;
    readonly isItalicText: boolean;
    isItalicTextFor(propName: string): boolean;
    readonly isPlacementCenter: boolean;
    isPlacementCenterFor(propName: string): boolean;
    readonly isPlacementLeft: boolean;
    isPlacementLeftFor(propName: string): boolean;
    readonly isPlacementRight: boolean;
    isPlacementRightFor(propName: string): boolean;
    readonly isPlacementStretchUnder: boolean;
    isPlacementStretchUnderFor(propName: string): boolean;
    readonly isPlacementUnder: boolean;
    isPlacementUnderFor(propName: string): boolean;
    readonly isUnderline: boolean;
    isUnderlineFor(propName: string): boolean;
    readonly objectId: string;
    readonly overrideText: string;
    overrideTextFor(propName: string): string;
    propAtIndex(index: number): Property;
    propAtName(propName: string): Property;
    readonly propCount: number;
    readonly propNames: Array<string>;
    readonly properties: Array<Property>;
    readonly propValues: Array<any>;
    readonly tipText: string;
    readonly type: string;
    tipTextFor(propName: string): string;
    toRecord(): Record;
    toJSON(): {
        id: string;
        properties: Property[];
        type: TypeNames;
    };
    valueAtName(propName: string): any;
}
export interface NullRedirection extends Redirection {
}
export declare class ObjectBinaryRef extends BinaryRef {
    constructor(settings: StringDictionary);
}
export declare class ObjectRef {
    readonly objectId: string;
    readonly description: string;
    readonly type: string;
    constructor(objectId: string, description: string, type: string);
    toString(): string;
}
export declare class PrintMarkup extends View {
    readonly cancelButtonText: string;
    readonly commitButtonText: string;
    readonly editable: boolean;
    readonly focusPropName: string;
    readonly printMarkupXML: string;
    readonly rows: Array<Array<Cell>>;
    private _orderedCellValue;
    readonly orderedCellValues: StringDictionary;
}
/**
 * Represents a 'value' or field in a row or record. See {@link Record}
 * A Prop has a corresponding {@link PropDef} that describes the property.
 * Like an {@link Record}, a Prop may also have {@link Annotation}s (style annotations),
 * but these apply to the property only
 */
export declare class Property {
    readonly name: string;
    readonly value: any;
    readonly propertyType: string;
    readonly format: string;
    readonly annotations: Array<DataAnnotation>;
    /**
     * Produce an unique string that can be used for comparison purposes
     * Props considered 'equal' should produce the same identity string
     *
     * @param o
     * @param propDef
     * @returns {any}
     */
    static identity(o: any, propDef: PropertyDef): string;
    static fromJSON(jsonObject: StringDictionary): Property;
    /**
     *
     * @param {string} name
     * @param value
     * @param {string} propertyType
     * @param {string} format
     * @param {Array<DataAnnotation>} annotations
     */
    constructor(name: string, value: any, propertyType?: string, format?: string, annotations?: Array<DataAnnotation>);
    equals(prop: Property): boolean;
    readonly backgroundColor: string;
    readonly foregroundColor: string;
    readonly imageName: string;
    readonly imagePlacement: string;
    readonly isBoldText: boolean;
    readonly isItalicText: boolean;
    readonly isPlacementCenter: boolean;
    readonly isPlacementLeft: boolean;
    readonly isPlacementRight: boolean;
    readonly isPlacementStretchUnder: boolean;
    readonly isPlacementUnder: boolean;
    readonly isUnderline: boolean;
    readonly overrideText: string;
    readonly tipText: string;
    readonly valueForWrite: any;
    toJSON(): {
        name: string;
        value: any;
        type: TypeNames;
    };
    private static parseValue(value, format);
}
/**
 /**
 * A property definition describes a particular value of a business entity. Business entities are transacted as records,
 * therefore properties and lists of properties are referred to as fields and records. Moreover, a list of property definitions
 * is referred to as a record definition and is the metadata describing the read/write capabilities of a specific dialog model
 * in use by a specific user in a specific workflow.
 * Contains information that 'defines' a property {@link Prop} (name/value)
 * An instance of the {@link Property} contains the actual data value.
 */
export declare class PropertyDef {
    /**
                 * The canCauseSideEffects meta property indicates that writing to this property can cause LOCAL side effects
                 * (on the same business object). For example, changing a 'zipCode' property case cause the 'state' property to change. If a user interface changes a property that can cause side effects, it should refresh the associated business view.
                 */ readonly canCauseSideEffects: boolean;
    readonly contentType: string;
    /**
     * Length of a type to be displayed. Some types are longer than what is practically needed by the application.
     * This property is used to define the practical length used in user interfaces.
     */
    readonly displayLength: number;
    /**
     * Scale of a decimal type to be displayed. Some decimal types are longer than what is practically needed by
     * the application. This property is used to define the practical scale used in user interfaces.
     */
    readonly displayScale: number;
    /**
     * The format property further describes the value using names that correlate, when possible, to the
     * Open API formats. Some example format names are date, date-time, uuid, int32 and int64. The format name decimal
     * is used to describe a string holding an arbitrary precision BCD value. For more information,
     * see the Open API Spec at https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md
     */
    readonly format: string;
    /**
     * Length of a type. This can be the length of a string or the length of a decimal.
     */
    readonly length: number;
    /**
     * The name of a business-data value
     */
    readonly propertyName: string;
    /**
     * Scale of a decimal type
     */
    readonly scale: number;
    /**
     * Whereas 'type' and 'format' define the physical meaning of a property, the 'semanticType' adds meaningful
     * insight into the usage of the property. For example, a 'decimal' type may be further defined semantically as a 'money' type.
     */
    readonly semanticType: string;
    /**
     * The type of business-data value. A valid type is either one of the OpenAPI basic values
     * [array boolean integer null number object string] or a fully qualified type name that can be retrieved from the schema service.
     */
    readonly propertyType: string;
    readonly writeAllowed: boolean;
    readonly writeEnabled: boolean;
    constructor(/**
                     * The canCauseSideEffects meta property indicates that writing to this property can cause LOCAL side effects
                     * (on the same business object). For example, changing a 'zipCode' property case cause the 'state' property to change. If a user interface changes a property that can cause side effects, it should refresh the associated business view.
                     */ canCauseSideEffects: boolean, contentType: string, 
        /**
         * Length of a type to be displayed. Some types are longer than what is practically needed by the application.
         * This property is used to define the practical length used in user interfaces.
         */
        displayLength: number, 
        /**
         * Scale of a decimal type to be displayed. Some decimal types are longer than what is practically needed by
         * the application. This property is used to define the practical scale used in user interfaces.
         */
        displayScale: number, 
        /**
         * The format property further describes the value using names that correlate, when possible, to the
         * Open API formats. Some example format names are date, date-time, uuid, int32 and int64. The format name decimal
         * is used to describe a string holding an arbitrary precision BCD value. For more information,
         * see the Open API Spec at https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md
         */
        format: string, 
        /**
         * Length of a type. This can be the length of a string or the length of a decimal.
         */
        length: number, 
        /**
         * The name of a business-data value
         */
        propertyName: string, 
        /**
         * Scale of a decimal type
         */
        scale: number, 
        /**
         * Whereas 'type' and 'format' define the physical meaning of a property, the 'semanticType' adds meaningful
         * insight into the usage of the property. For example, a 'decimal' type may be further defined semantically as a 'money' type.
         */
        semanticType: string, 
        /**
         * The type of business-data value. A valid type is either one of the OpenAPI basic values
         * [array boolean integer null number object string] or a fully qualified type name that can be retrieved from the schema service.
         */
        propertyType: string, writeAllowed: boolean, writeEnabled: boolean);
    readonly isBarcodeType: boolean;
    readonly isBinaryType: boolean;
    readonly isBooleanType: boolean;
    readonly isCodeRefType: boolean;
    readonly isDateType: boolean;
    readonly isDateTimeType: boolean;
    readonly isDecimalType: boolean;
    readonly isDoubleType: boolean;
    readonly isEmailType: boolean;
    readonly isFileAttachment: boolean;
    readonly isFloatType: boolean;
    readonly isGpsReadingType: boolean;
    readonly isMapLocationType: boolean;
    readonly isHTMLType: boolean;
    readonly isInlineMediaStyle: boolean;
    readonly isListType: boolean;
    readonly isIntType: boolean;
    readonly isLargeBinaryType: boolean;
    readonly isLongType: boolean;
    readonly isMoneyType: boolean;
    readonly isNumericType: boolean;
    readonly isObjRefType: boolean;
    readonly isPasswordType: boolean;
    readonly isPercentType: boolean;
    readonly isSignatureType: boolean;
    readonly isStringType: boolean;
    readonly isTelephoneType: boolean;
    readonly isTextBlock: boolean;
    readonly isTimeType: boolean;
    readonly isUnformattedNumericType: boolean;
    readonly isURLType: boolean;
}
export declare class PropFormatter {
    static decimalFormat: string[];
    static decimalFormatGeneric: string;
    static moneyFormat: string[];
    static moneyFormatGeneric: string;
    static percentFormat: string[];
    static percentFormatGeneric: string;
    static wholeFormat: string;
    /**
     * Get a string representation of this property suitable for 'reading'
     * @param prop
     * @param propDef
     * @returns {string}
     */
    static formatForRead(prop: Property, propDef: PropertyDef): string;
    static formatValueForRead(value: any, propDef: PropertyDef): any;
    /**
     * Get a string representation of this property suitable for 'writing'
     * @param prop
     * @param propDef
     * @returns {string}
     */
    static formatForWrite(prop: Property, propDef: PropertyDef): string;
    /**
     * Attempt to construct (or preserve) the appropriate data type given primitive (or already constructed) value.
     * @param value
     * @param propDef
     * @returns {any}
     */
    static parse(value: any, propDef: PropertyDef): any;
    static resetFormats(): void;
    static toString(o: any, propDef: PropertyDef): string;
    /**
     * Render this value as a string
     * @param o
     * @param propDef
     * @returns {any}
     */
    static toStringRead(o: any, propDef: PropertyDef): string;
    static toStringWrite(o: any, propDef: PropertyDef): string;
}
export interface QueryParameters {
    fetchDirection: QueryDirection;
    fetchMaxRecords: number;
    fromBusinessId?: string;
    type: string;
}
export interface Record {
    readonly annotations?: Array<DataAnnotation>;
    readonly id: string;
    properties: Array<Property>;
    type: string;
    annotationsAtName(propName: string): Array<DataAnnotation>;
    afterEffects(after: Record): Record;
    backgroundColor: string;
    backgroundColorFor(propName: string): string;
    foregroundColor: string;
    foregroundColorFor(propName: string): string;
    imageName: string;
    imageNameFor(propName: string): string;
    imagePlacement: string;
    imagePlacementFor(propName: string): string;
    isBoldText: boolean;
    isBoldTextFor(propName: string): boolean;
    isItalicText: boolean;
    isItalicTextFor(propName: string): boolean;
    isPlacementCenter: boolean;
    isPlacementCenterFor(propName: string): boolean;
    isPlacementLeft: boolean;
    isPlacementLeftFor(propName: string): boolean;
    isPlacementRight: boolean;
    isPlacementRightFor(propName: string): boolean;
    isPlacementStretchUnder: boolean;
    isPlacementStretchUnderFor(propName: string): boolean;
    isPlacementUnder: boolean;
    isPlacementUnderFor(propName: string): boolean;
    isUnderline: boolean;
    isUnderlineFor(propName: string): boolean;
    overrideText: string;
    overrideTextFor(propName: string): string;
    propAtIndex(index: number): Property;
    propAtName(propName: string): Property;
    propCount: number;
    propNames: Array<string>;
    propValues: Array<any>;
    tipText: string;
    tipTextFor(propName: string): string;
    toJSON(): any;
    valueAtName(propName: string): any;
}
/**
 * In the same way that a {@link PropertyDef} describes a {@link Property}, a RecordDef describes an {@link Record}.
 * It is composed of {@link PropertyDef}s while the {@link Record} is composed of {@link Property}s.
 * In other words it describes the structure or makeup of a row or record, but does not contain the data values themselves.
 * The corresponding {@link Record} contains the actual values.
 */
export declare class RecordDef {
    readonly propertyDefs: Array<PropertyDef>;
    readonly propCount: number;
    propDefAtName(name: string): PropertyDef;
    readonly propNames: Array<string>;
}
export interface RecordSet {
    defaultActionId: string;
    hasMore: boolean;
    records: Array<Record>;
}
export interface Redirection {
    readonly id: string;
    readonly referringObject: ReferringObject;
    readonly sessionId: string;
    readonly tenantId: string;
    readonly type: RedirectionType;
}
export declare class RedirectionUtil {
    static isRedirection(o: any): boolean;
    static isDialogRedirection(o: any): boolean;
    static isNullRedirection(o: any): boolean;
    static isWebRedirection(o: any): boolean;
    static isWorkbenchRedirection(o: any): boolean;
}
export interface Session {
    readonly appVendors: ReadonlyArray<string>;
    /**
     * Current version of the underlying application (business) logic. This is not the middleware version.
     */
    readonly appVersion: string;
    readonly appWindow: AppWindow;
    /**
     * Current division is analagous to a \"sub-tenant\" and it's possible for users to tailor different desktops based
     * on their division.
     */
    readonly currentDivision: string;
    readonly id: string;
    readonly tenantId: string;
    /**
     * The dialog layer interacts with an application endpoint, transparent to the user interface. The serverAssignment
     * is not used by the user interface directly, but this value can be useful when diagnosing problems.
     */
    readonly serverAssignment: string;
    /**
     * Current version of the underlying middleware (not the application logic)
     */
    readonly serverVersion: string;
    /**
     * The tenantProperties object is arbitrary, its values are dynamically defined and it is used to return
     * tenant-specific values to the requesting client.
     */
    readonly tenantProperties: StringDictionary;
    readonly type: string;
    readonly userId: string;
}
export interface Sort {
    readonly propertyName: string;
    readonly direction: SortDirection;
}
export declare class Stream extends View {
    readonly topic: string;
    readonly bufferSize: number;
    readonly view: View;
}
/**
 * A text template containing substitution parameters that is instantiated at presentation time and filled with business values.
 */
export declare class SubstitutionCellValue extends CellValue {
    readonly value: string;
    constructor(style: any, value: string);
}
/**
 * A purely declarative type. This object has no additional properties.
 */
export declare class TabCellValue extends CellValue {
    constructor(style: any);
}
export interface Tenant {
    readonly id: string;
    readonly description: string;
}
export declare class UserMessage {
    message: string;
    messageType: string;
    explanation: string;
    propertyNames: Array<string>;
    constructor(message: string, messageType: string, explanation: string, propertyNames: Array<string>);
}
export declare class ViewDescriptor {
    readonly id: string;
    readonly name: string;
    readonly title: string;
    constructor(id: string, name: string, title: string);
}
export interface Workbench {
    readonly actions: ReadonlyArray<WorkbenchAction>;
    readonly id: string;
    readonly name: string;
    readonly offlineCapable: boolean;
}
export interface WorkbenchAction {
    /**
     * An alternative unique identifier. An alias is typically used to identify an action that is used in offline mode.
     * An alias should be more descriptive than the id, therefore an alias makes offline logic more descriptive. descriptive.
     */
    readonly actionId?: string;
    readonly alias: string;
    readonly id: string;
    readonly iconBase: string;
    readonly name: string;
    readonly workbenchId: string;
}
export interface NullRedirection extends Redirection {
}
export interface WebRedirection extends Redirection {
    readonly url: string;
}
export interface WorkbenchRedirection extends Redirection {
    readonly workbenchId: string;
}
/**
 * Top-level class, representing a Catavolt 'Dialog' definition.
 * All Dialogs have a composite {@link View} definition along with a single record
 * or a list of records.  See {@Record}
 */
export declare abstract class Dialog {
    static BINARY_CHUNK_SIZE: number;
    private static CHAR_CHUNK_SIZE;
    private _binaryCache;
    private _lastRefreshTime;
    private _catavolt;
    protected _parentDialog: any;
    readonly availableViews: Array<ViewDescriptor>;
    readonly businessClassName: string;
    readonly children: Array<Dialog>;
    readonly description: string;
    readonly dialogClassName: string;
    dialogMode: DialogMode;
    readonly header: View;
    readonly id: string;
    readonly recordDef: RecordDef;
    readonly referringObject: ReferringObject;
    readonly selectedViewId: string;
    readonly sessionId: string;
    readonly tenantId: string;
    readonly type: DialogType;
    readonly view: View;
    readonly viewMode: ViewMode;
    readonly catavolt: CatavoltApi;
    /**
     * Load a Binary property from a record
     * @param propName
     * @param record
     * @returns {any}
     */
    binaryAt(propName: string, record: Record): Promise<Binary>;
    destroy(): void;
    /**
     * Return the error associated with this dialog, if any
     * @returns {any}
     */
    readonly error: DialogException;
    /**
     * Find a menu def on this dialog with the given actionId
     * @param actionId
     * @returns {Menu}
     */
    findMenuAt(actionId: string): Menu;
    /**
     * Get a string representation of this property suitable for 'reading'
     * @param propValue
     * @param propName
     * @returns {string}
     */
    formatForRead(prop: Property, propName: string): string;
    /**
     * Get a string representation of this property suitable for 'writing'
     * @param propValue
     * @param propName
     * @returns {string}
     */
    formatForWrite(prop: Property, propName: string): string;
    /**
     * Returns whether or not this dialog loaded properly
     * @returns {boolean}
     */
    readonly hasError: boolean;
    /**
     * Returns whether or not this Form is destroyed
     * @returns {boolean}
     */
    readonly isDestroyed: boolean;
    /**
     * Returns whether or not the data in this dialog is out of date
     * @returns {boolean}
     */
    readonly isRefreshNeeded: boolean;
    /**
     * Get the last time this dialog's data was refreshed
     * @returns {Date}
     */
    /**
     * @param time
     */
    lastRefreshTime: Date;
    /**
     * Get the all {@link Menu}'s associated with this dialog
     * @returns {Array<Menu>}
     */
    readonly menu: Menu;
    openView(targetViewDescriptor: ViewDescriptor): Promise<Dialog>;
    /**
     * Get the title of this dialog
     * @returns {string}
     */
    readonly paneTitle: string;
    /**
     * Parses a value to prepare for 'writing' back to the server
     * @param formattedValue
     * @param propName
     * @returns {any}
     */
    parseValue(formattedValue: any, propName: string): any;
    /**
     * Get the propery definition for a property name
     * @param propName
     * @returns {PropDef}
     */
    propDefAtName(propName: string): PropertyDef;
    /**
     * Read all the Binary values in this {@link Record}
     * @param record
     * @returns {Future<Array<Try<Binary>>>}
     */
    readBinaries(record: Record): Promise<Array<Binary>>;
    readonly parentDialog: Dialog;
    /**
     * Get the all {@link ViewDescriptor}'s associated with this Form
     * @returns {Array<ViewDescriptor>}
     */
    readonly viewDescs: Array<ViewDescriptor>;
    writeAttachment(attachment: Attachment): Promise<void>;
    writeAttachments(record: Record): Promise<Array<void>>;
    /**
     * Write all Binary values in this {@link Record} back to the server
     * @param record
     * @returns {Future<Array<Try<XWritePropertyResult>>>}
     */
    writeBinaries(record: Record): Promise<Array<void>>;
    initialize(catavolt: CatavoltApi): void;
    /**
     * Perform this action associated with the given Menu on this dialog.
     * The targets array is expected to be an array of object ids.
     * @param {Menu} menu
     * @param {ActionParameters} actionParams
     * @returns {Promise<{actionId: string} | Redirection>}
     */
    protected invokeMenuAction(menu: Menu, actionParams: ActionParameters): Promise<{
        actionId: string;
    } | Redirection>;
    protected readBinary(propName: string, record: Record): Promise<Binary>;
    protected updateSettingsWithNewDialogProperties(referringObject: ReferringObject): void;
    /**
     * @private
     * @returns {boolean}
     */
    private readonly isAnyChildDestroyed;
}
/**
 * PanContext Subtype that represents an 'Editor Dialog'.
 * An 'Editor' represents and is backed by a single Record and Record definition.
 * See {@link Record} and {@link RecordDef}.
 */
export declare class EditorDialog extends Dialog {
    private _buffer;
    private _isFirstReadComplete;
    readonly businessId: string;
    /**
     * Get the current buffered record
     * @returns {RecordBuffer}
     */
    readonly buffer: RecordBuffer;
    changeViewMode(viewMode: ViewMode): Promise<EditorDialog>;
    /**
     * Get the associated entity record
     * @returns {Record}
     */
    readonly record: Record;
    /**
     * Get the current version of the entity record, with any pending changes present
     * @returns {Record}
     */
    readonly recordNow: Record;
    /**
     * Get the possible values for a 'constrained value' property
     * @param propName
     * @returns {Future<Array<any>>}
     */
    getAvailableValues(propName: string): Promise<Array<Object>>;
    /**
     * Returns whether or not this cell definition contains a binary value
     * @param cellValueDef
     * @returns {PropDef|boolean}
     */
    isBinary(cellValue: AttributeCellValue): boolean;
    /**
     * Returns whether or not the buffers contain valid data via a successful read operation.
     * @returns {boolean}
     */
    readonly isFirstReadComplete: boolean;
    /**
     * Returns whether or not this Editor is in 'read' mode
     * @returns {boolean}
     */
    readonly isReadMode: boolean;
    /**
     * Returns whether or not this property is read-only
     * @param propName
     * @returns {boolean}
     */
    isReadModeFor(propName: string): boolean;
    /**
     * Returns whether or not this cell definition contains a binary value that should be treated as a signature control
     * @param cellValueDef
     * @returns {PropDef|boolean}
     */
    isSignature(cellValueDef: AttributeCellValue): boolean;
    /**
     * Returns whether or not this property is 'writable'
     * @returns {boolean}
     */
    readonly isWriteMode: boolean;
    /**
     * Perform the action associated with the given Menu on this EditorDialog
     * Given that the Editor could possibly be destroyed as a result of this action,
     * any provided pending writes will be saved if present.
     * @param {Menu} menu
     * @param {Record} pendingWrites
     * @returns {Promise<{actionId: string} | Redirection>}
     */
    performMenuAction(menu: Menu, pendingWrites: Record): Promise<{
        actionId: string;
    } | Redirection>;
    /**
     * Properties whose {@link PropDef.canCauseSideEffects} value is true, may change other underlying values in the model.
     * This method will update those underlying values, given the property name that is changing, and the new value.
     * This is frequently used with {@link EditorContext.getAvailableValues}.  When a value is seleted, other properties
     * available values may change. (i.e. Country, State, City dropdowns)
     * @param propertyName
     * @param value
     * @returns {Future<null>}
     */
    processSideEffects(propertyName: string, value: any): Promise<void>;
    /**
     * Read (load) the {@link Record} assocated with this Editor
     * The record must be read at least once to initialize the Context
     * @returns {Future<Record>}
     */
    read(): Promise<Record>;
    /**
     * Set the value of a property in this {@link Record}.
     * Values may be already constructed target types (CodeRef, TimeValue, Date, etc.)
     * or primitives, in which case the values will be parsed and objects constructed as necessary.
     * @param name
     * @param value
     * @returns {any}
     */
    setPropValue(name: string, value: any): any;
    /**
     * Set a binary property from a string formatted as a 'data url'
     * See {@link https://en.wikipedia.org/wiki/Data_URI_scheme}
     * @param name
     * @param dataUrl
     */
    setBinaryPropWithDataUrl(name: string, dataUrl: string): void;
    /**
     * Set a binary property with base64 encoded data
     * @param name
     * @param encodedData
     * @param mimeType
     */
    setBinaryPropWithEncodedData(name: string, encodedData: string, mimeType: string): void;
    /**
     * Write this record (i.e. {@link Record}} back to the server
     * @returns {Promise<Record | Redirection>}
     */
    write(): Promise<Record | Redirection>;
    private removeSpecialProps(record);
    private initBuffer(record);
}
/**
 * Dialog Subtype that represents a 'Query Dialog'.
 * A 'Query' represents and is backed by a list of Records and a single Record definition.
 * See {@link Record} and {@link RecordDef}.
 */
export declare class QueryDialog extends Dialog {
    private _scroller;
    private _defaultActionId;
    positionalQueryAbility: PositionalQueryAbilityType;
    supportsColumnStatistics: boolean;
    supportsPositionalQueries: boolean;
    /**
     * Returns whether or not a column is of a binary type
     * @param columnDef
     * @returns {PropDef|boolean}
     */
    isBinary(column: Column): boolean;
    readonly defaultActionId: string;
    /**
     * Perform this action associated with the given Menu on this dialog.
     * The targets array is expected to be an array of object ids.
     * @param {Menu} menu
     * @param {Array<string>} targets
     * @returns {Promise<{actionId: string} | Redirection>}
     */
    performMenuAction(menu: Menu, targets: Array<string>): Promise<{
        actionId: string;
    } | Redirection>;
    /**
     * Perform a query
     * Note: {@link QueryScroller} is the preferred way to perform a query.
     * see {@link QueryContext.newScroller} and {@link QueryContext.setScroller}
     * @param maxRows
     * @param direction
     * @param fromObjectId
     * @returns {Future<RecordSet>}
     */
    query(maxRows: number, direction: QueryDirection, fromObjectId: string): Promise<RecordSet>;
    /**
     * Clear the QueryScroller's buffer and perform this query
     * @returns {Future<Array<Record>>}
     */
    refresh(): Promise<Array<Record>>;
    /**
     * Get the associated QueryScroller
     * @returns {QueryScroller}
     */
    readonly scroller: QueryScroller;
    /**
     * Creates a new QueryScroller with the given values
     * @param pageSize
     * @param firstObjectId
     * @param markerOptions
     * @returns {QueryScroller}
     */
    setScroller(pageSize: number, firstObjectId?: string, markerOptions?: Array<QueryMarkerOption>): QueryScroller;
    /**
     * Creates a new QueryScroller with default buffer size of 50
     * @returns {QueryScroller}
     */
    newScroller(): QueryScroller;
}
/**
 * *********************************
 */
export declare class HasMoreQueryMarker extends NullRecord {
    static singleton: HasMoreQueryMarker;
}
export declare class IsEmptyQueryMarker extends NullRecord {
    static singleton: IsEmptyQueryMarker;
}
export declare enum QueryMarkerOption {
    None = 0,
    IsEmpty = 1,
    HasMore = 2,
}
export declare class QueryScroller {
    private _dialog;
    private _pageSize;
    private _firstObjectId;
    private _markerOptions;
    private _buffer;
    private _hasMoreBackward;
    private _hasMoreForward;
    private _nextPagePromise;
    private _prevPagePromise;
    private _firstResultOid;
    constructor(_dialog: QueryDialog, _pageSize: number, _firstObjectId: string, _markerOptions?: Array<QueryMarkerOption>);
    readonly buffer: Array<Record>;
    readonly bufferWithMarkers: Array<Record>;
    readonly dialog: QueryDialog;
    readonly firstObjectId: string;
    readonly hasMoreBackward: boolean;
    readonly hasMoreForward: boolean;
    readonly isComplete: boolean;
    readonly isCompleteAndEmpty: boolean;
    readonly isEmpty: boolean;
    pageBackward(): Promise<Array<Record>>;
    pageForward(): Promise<Array<Record>>;
    readonly pageSize: number;
    refresh(): Promise<Array<Record>>;
    trimFirst(n: number): void;
    trimLast(n: number): void;
    private clear();
}
export declare class ErrorDialog extends Dialog {
}
export declare type AttributeCellValueEntryMethod = "COMBO_BOX" | "DROP_DOWN" | "TEXT_FIELD" | "ICON_CHOOSER";
export declare type ClientType = 'DESKTOP' | 'MOBILE';
export declare type DialogMessageMessageType = "CONFIRM" | "ERROR" | "INFO" | "WARN";
export declare enum DialogModeEnum {
    COPY = "COPY",
    CREATE = "CREATE",
    READ = "READ",
    UPDATE = "UPDATE",
    DESTROYED = "DESTROYED",
    DELETE = "DELETE",
    LIST = "LIST",
}
export declare type DialogMode = DialogModeEnum.COPY | DialogModeEnum.CREATE | DialogModeEnum.READ | DialogModeEnum.UPDATE | DialogModeEnum.DESTROYED | DialogModeEnum.DELETE | DialogModeEnum.LIST;
export declare type DialogType = 'hxgn.api.dialog.EditorDialog' | 'hxgn.api.dialog.QueryDialog';
export declare type FilterOperator = "AND" | "CONTAINS" | "ENDS_WITH" | "EQUAL_TO" | "GREATER_THAN" | "GREATER_THAN_OR_EQUAL_TO" | "LESS_THAN" | "LESS_THAN_OR_EQUAL_TO" | "NOT_EQUAL_TO" | "OR" | "STARTS_WITH";
export declare type PositionalQueryAbilityType = 'FULL' | 'NONE';
export declare enum QueryDirectionEnum {
    FORWARD = "FORWARD",
    BACKWARD = "BACKWARD",
}
export declare type QueryDirection = QueryDirectionEnum.FORWARD | QueryDirectionEnum.BACKWARD;
export declare type RedirectionType = 'hxgn.api.dialog.DialogRedirection' | 'hxgn.api.dialog.WebRedirection' | 'hxgn.api.dialog.WorkbenchRedirection' | 'hxgn.api.dialog.NullRedirection';
export declare type SortDirection = "ASC" | "DESC";
export declare enum ViewModeEnum {
    READ = "READ",
    WRITE = "WRITE",
}
export declare type ViewMode = ViewModeEnum.READ | ViewModeEnum.WRITE;
export declare type ViewType = 'hxgn.api.dialog.BarcodeScan' | 'hxgn.api.dialog.Calendar' | 'hxgn.api.dialog.Details' | 'hxgn.api.dialog.Form' | 'hxgn.api.dialog.GpsReading' | 'hxgn.api.dialog.MapLocation' | 'hxgn.api.dialog.Graph' | 'hxgn.api.dialog.List' | 'hxgn.api.dialog.Map' | 'hxgn.api.dialog.Stream';
export declare enum TypeNames {
    ActionParametersTypeName = "hxgn.api.dialog.ActionParameters",
    AppWindowTypeName = "hxgn.api.dialog.AppWindow",
    BarcodeScan = "hxgn.api.dialog.BarcodeScan",
    CalendarTypeName = "hxgn.api.dialog.Calendar",
    CodeRefTypeName = "hxgn.api.dialog.CodeRef",
    DetailsTypeName = "hxgn.api.dialog.Details",
    DialogTypeName = "hxgn.api.dialog.Dialog",
    DialogMessageTypeName = "hxgn.api.dialog.DialogMessage",
    DialogRedirectionTypeName = "hxgn.api.dialog.DialogRedirection",
    EditorDialogTypeName = "hxgn.api.dialog.EditorDialog",
    FormTypeName = "hxgn.api.dialog.Form",
    GpsReadingTypeName = "hxgn.api.dialog.GpsReading",
    GpsReadingPropertyTypeName = "hxgn.api.dialog.GpsReadingProperty",
    GraphTypeName = "hxgn.api.dialog.Graph",
    ListTypeName = "hxgn.api.dialog.List",
    LoginTypeName = "hxgn.api.dialog.Login",
    MapTypeName = "hxgn.api.dialog.Map",
    MapLocationTypeName = "hxgn.api.dialog.MapLocation",
    MapLocationPropertyTypeName = "hxgn.api.dialog.MapLocationProperty",
    NullRedirectionTypeName = "hxgn.api.dialog.NullRedirection",
    ObjectRefTypeName = "hxgn.api.dialog.ObjectRef",
    PropertyTypeName = "hxgn.api.dialog.Property",
    QueryDialogTypeName = "hxgn.api.dialog.QueryDialog",
    QueryParametersTypeName = "hxgn.api.dialog.QueryParameters",
    RecordTypeName = "hxgn.api.dialog.Record",
    ReferringDialogTypeName = "hxgn.api.dialog.ReferringDialog",
    ReferringWorkbenchTypeName = "hxgn.api.dialog.ReferringWorkbench",
    SessionTypeName = "hxgn.api.dialog.Session",
    StreamTypeName = "hxgn.api.dialog.Stream",
    WebRedirectionTypeName = "hxgn.api.dialog.WebRedirection",
    WorkbenchTypeName = "hxgn.api.dialog.Workbench",
    WorkbenchRedirectionTypeName = "hxgn.api.dialog.WorkbenchRedirection",
}
export declare class ModelUtil {
    private static classTypes;
    private static classType(name);
    private static typeInstance(name);
    static jsonToModel<A>(obj: any, n?: number): Promise<A>;
    static modelToJson(obj: any, filterFn?: (prop) => boolean): StringDictionary;
    private static deserializeArray(array);
    private static assignProp(prop, value, target, type, n);
}
